"""
.. module:: array_manager
   :synopsis: Extendible NumPy array manager
.. moduleauthor:: David Bindel <bindel@cornell.edu>

:Module: array_manager
:Author: David Bindel <bindel@cornell.edu>
"""


import numpy as np


def extend_size(d, dmax):
    """Determine size to allocate on resize operation.

    Args:
        d: Desired length
        dmax: Current allocation size

    Returns:
        New allocation size
    """
    if d <= dmax:
        return dmax
    else:
        return 2*d


class ArrayManager(object):
    """Manage extendible NumPy array storage.

    An ArrayManager handles copying and reallocation for an extensible
    storage block (as a NumPy array) attached to a NumPy view object.
    The manager does not ask as a proxy for the view.

    Attributes:
        view: View on the current active array
        shape: Current array shape
        maxdims: Maximum dimensions allowed without reallocation
    """

    def __init__(self, dims, maxdims=None, **kwargs):
        if maxdims is None:
            maxdims = dims
        self._init_kwargs = kwargs
        self._dims = dims
        self._data = np.zeros(maxdims, **self._init_kwargs)
        self._view = self._make_view()

    def _make_view(self):
        return self._data[tuple(slice(0, d) for d in self._dims)]

    @property
    def maxdims(self):
        return self._data.shape

    @property
    def shape(self):
        return self._dims

    @shape.setter
    def shape(self, newdims):
        self.resize(newdims)

    @property
    def view(self):
        return self._view

    def resize(self, dims):
        "Resize the array to the requested size"
        dnew = [extend_size(d, dmax)
                for d, dmax in zip(dims, self._data.shape)]
        if dnew != self._data.shape:
            data = np.zeros(dnew, **self._init_kwargs)
            slices = tuple(slice(0, d) for d in self._dims)
            data[slices] = self._data[slices]
            self._data = data
            self._view = self._data[slices]
        self._dims = tuple(dims)
        self._view = self._make_view()

    def append(self, data, axis=0):
        "Add data along a given axis"
        if not isinstance(data, np.ndarray):
            data = np.array(data)
        if len(data.shape) == len(self._dims)-1:
            data = np.expand_dims(data, axis=axis)
        daxis0 = self._dims[axis]
        daxis1 = daxis0 + data.shape[axis]
        newdims = list(self._dims)
        newdims[axis] = daxis1
        self.resize(newdims)
        slices = [slice(0, d) for d in self._dims]
        slices[axis] = slice(daxis0, daxis1)
        self._data[tuple(slices)] = data


def test_array_manager():
    "Test behavior of ArrayManager"

    mgr = ArrayManager((2, 3))
    v = mgr.view
    v[0, 0] = 1
    v[1, 2] = 2
    assert (v == np.array([[1, 0, 0],
                           [0, 0, 2]])).all(), "Wrong initial view"
    mgr.shape = (3, 3)
    v2 = mgr.view
    assert (v2 == np.array([[1, 0, 0],
                            [0, 0, 2],
                            [0, 0, 0]])).all(), "Wrong second view"

    assert mgr.maxdims[0] == 6, "Resize not as expected"
    assert mgr.maxdims[1] == 3, "Resize not as expected"

    mgr.append(np.array([1, 1, 1]))
    v3 = mgr.view
    assert (v3 == np.array([[1, 0, 0],
                            [0, 0, 2],
                            [0, 0, 0],
                            [1, 1, 1]])).all(), "Wrong third view"

    assert mgr.maxdims[0] == 6, "Resize not as expected"
    assert mgr.maxdims[1] == 3, "Resize not as expected"

    mgr.append(np.array([[1], [2], [3], [4]]), axis=1)
    v4 = mgr.view
    assert (v4 == np.array([[1, 0, 0, 1],
                            [0, 0, 2, 2],
                            [0, 0, 0, 3],
                            [1, 1, 1, 4]])).all(), "Wrong fourth view"

    assert mgr.maxdims[0] == 6, "Resize not as expected"
    assert mgr.maxdims[1] == 8, "Resize not as expected"

    mgr2 = ArrayManager((5,), dtype=np.int)
    assert mgr2.view.dtype == np.int, "dtype transfer fails"

    mgr2.append(np.array([[1]]))
    assert mgr2.view.dtype == np.int, "dtype transfer fails"


if __name__ == "__main__":
    test_array_manager()



"""
.. module:: rs_capped
   :synopsis: Median-capped interpolation
.. moduleauthor:: David Bindel <bindel@cornell.edu>

:Module: rs_capped
:Author: David Bindel <bindel@cornell.edu>
"""

import numpy as np
from utils import from_unit_box, to_unit_box


class RSCapped(object):
    """Cap adapter for RBF response surface.

    This adapter takes an existing response surface and replaces it
    with a modified version in which any function values above the
    median are replaced by the median value.

    :ivar model: Original response surface
    :ivar fvalues: Function values
    """

    def __init__(self, model, transformation=None):
        """Initialize the response surface adapter

        :param model: Original response surface object
        :param transformation: Function value transformation
        """
        self.needs_update = False
        self.transformation = transformation
        if self.transformation is None:
            def transformation(fvalues):
                medf = np.median(fvalues)
                fvalues[fvalues > medf] = medf
                return fvalues
            self.transformation = transformation
        self.model = model
        self.fvalues = np.zeros((model.maxp, 1))
        self.nump = 0
        self.maxp = model.maxp

    @property
    def x(self):
        return self.get_x()

    @property
    def fx(self):
        return self.get_fx()

    def reset(self):
        """Reset the capped response surface
        """
        self.model.reset()
        self.fvalues[:] = 0
        self.nump = 0

    def add_point(self, xx, fx):
        """Add a new function evaluation

        :param xx: Point to add
        :param fx: The function value of the point to add
        """
        if self.nump >= self.fvalues.shape[0]:
            self.fvalues.resize(2*self.fvalues.shape[0], 1)
        self.fvalues[self.nump] = fx
        self.nump += 1
        self.needs_update = True
        self.model.add_point(xx, fx)

    def get_x(self):
        """Get the list of data points

        :return: List of data points
        """
        return self.model.get_x()

    def get_fx(self):
        """Get the list of function values for the data points.

        :return: List of function values
        """
        return self.model.get_fx()

    def eval(self, xx, d=None):
        """Evaluate the capped rbf interpolant at the point xx

        :param xx: Point where to evaluate
        :return: Value of the capped rbf interpolant at x
        """
        self._apply_transformation()
        return self.model.eval(xx, d)

    def evals(self, xx, d=None):
        """Evaluate the capped rbf interpolant at the points xx

        :param xx: Points where to evaluate
        :return: Values of the capped rbf interpolant at x
        """
        self._apply_transformation()
        return self.model.evals(xx, d)

    def deriv(self, xx, d=None):
        """Evaluate the derivative of the rbf interpolant at x

        :param x: Data point
        :return: Derivative of the rbf interpolant at x
        """
        self._apply_transformation()
        return self.model.deriv(xx, d)

    def _apply_transformation(self):
        """ Apply the cap to the function values.
        """
        fvalues = np.copy(self.fvalues[0:self.nump])
        self.model.transform_fx(self.transformation(fvalues))


class RSPenalty(object):
    """Cap adapter for RBF response surface.

    This adapter takes an existing response surface and replaces it
    with a modified version in which any function values above the
    median are replaced by the median value.

    :ivar model: Original response surface
    :ivar fvalues: Function values
    """

    def __init__(self, model, evals, derivs):
        """Initialize the response surface adapter

        :param model: Original response surface object
        """
        self.needs_update = False
        self.model = model
        self.fvalues = np.zeros((model.maxp, 1))
        self.nump = 0
        self.maxp = model.maxp
        self.eval_method = evals
        self.deriv_method = derivs

    @property
    def x(self):
        return self.get_x()

    @property
    def fx(self):
        return self.get_fx()

    def reset(self):
        """Reset the capped response surface
        """
        self.model.reset()
        self.fvalues[:] = 0
        self.nump = 0

    def add_point(self, xx, fx):
        """Add a new function evaluation

        :param xx: Point to add
        :param fx: The function value of the point to add
        """
        if self.nump >= self.fvalues.shape[0]:
            self.fvalues.resize(2*self.fvalues.shape[0], 1)
        self.fvalues[self.nump] = fx
        self.nump += 1
        self.needs_update = True
        self.model.add_point(xx, fx)

    def get_x(self):
        """Get the list of data points

        :return: List of data points
        """
        return self.model.get_x()

    def get_fx(self):
        """Get the list of function values for the data points.

        :return: List of function values
        """
        return self.eval_method(self.model, self.model.get_x())[0, 0]

    def eval(self, xx, d=None):
        """Evaluate the capped rbf interpolant at the point xx

        :param xx: Point where to evaluate
        :return: Value of the capped rbf interpolant at x
        """
        return self.eval_method(self.model, np.atleast_2d(xx)).ravel()

    def evals(self, xx, d=None):
        """Evaluate the capped rbf interpolant at the points xx

        :param xx: Points where to evaluate
        :return: Values of the capped rbf interpolant at x
        """

        return self.eval_method(self.model, xx)

    def deriv(self, xx, d=None):
        """Evaluate the derivative of the rbf interpolant at x

        :param x: Data point
        :return: Derivative of the rbf interpolant at x
        """

        return self.deriv_method(self.model, xx)


class RSUnitbox(object):
    """Cap adapter for RBF response surface.

    This adapter takes an existing response surface and replaces it
    with a modified version in which any function values above the
    median are replaced by the median value.

    :ivar model: Original response surface
    :ivar fvalues: Function values
    """

    def __init__(self, model, data):
        """Initialize the response surface adapter

        :param model: Original response surface object
        """
        self.needs_update = False
        self.model = model
        self.fvalues = np.zeros((model.maxp, 1))
        self.nump = 0
        self.maxp = model.maxp
        self.data = data

    @property
    def x(self):
        return self.get_x()

    @property
    def fx(self):
        return self.get_fx()

    def reset(self):
        """Reset the capped response surface
        """
        self.model.reset()
        self.fvalues[:] = 0
        self.nump = 0

    def add_point(self, xx, fx):
        """Add a new function evaluation

        :param xx: Point to add
        :param fx: The function value of the point to add
        """
        if self.nump >= self.fvalues.shape[0]:
            self.fvalues.resize(2*self.fvalues.shape[0], 1)
        self.fvalues[self.nump] = fx
        self.nump += 1
        self.needs_update = True
        self.model.add_point(to_unit_box(xx, self.data), fx)

    def get_x(self):
        """Get the list of data points

        :return: List of data points
        """
        return from_unit_box(self.model.get_x(), self.data)

    def get_fx(self):
        """Get the list of function values for the data points.

        :return: List of function values
        """
        return self.model.get_fx()

    def eval(self, xx, d=None):
        """Evaluate the capped rbf interpolant at the point xx

        :param xx: Point where to evaluate
        :return: Value of the capped rbf interpolant at x
        """
        return self.model.eval(to_unit_box(xx, self.data))

    def evals(self, xx, d=None):
        """Evaluate the capped rbf interpolant at the points xx

        :param xx: Points where to evaluate
        :return: Values of the capped rbf interpolant at x
        """

        return self.model.evals(to_unit_box(xx, self.data))

    def deriv(self, xx, d=None):
        """Evaluate the derivative of the rbf interpolant at x

        :param x: Data point
        :return: Derivative of the rbf interpolant at x
        """

        return self.model.deriv(to_unit_box(xx, self.data))



"""
.. module:: test_simple
  :synopsis: Test Simple
.. moduleauthor:: David Eriksson <dme65@cornell.edu>
"""

from pySOT import *
from poap.controller import ThreadController, BasicWorkerThread
import numpy as np
import os.path
import time


def main():
    if not os.path.exists("./logfiles"):
        os.makedirs("logfiles")
    if os.path.exists("./logfiles/test_simple.log"):
        os.remove("./logfiles/test_simple.log")
    logging.basicConfig(filename="./logfiles/test_simple.log",
                        level=logging.INFO)

    print("\nNumber of threads: 4")
    print("Maximum number of evaluations: 1000")
    print("Search strategy: CandidateDYCORS")
    print("Experimental design: Latin Hypercube")
    print("Surrogate: Cubic RBF, domain scaled to unit box")

    nthreads = 4
    maxeval = 500
    nsamples = nthreads

    data = Ackley(dim=10)
    print(data.info)

    # Create a strategy and a controller
    controller = ThreadController()
    controller.strategy = \
        SyncStrategyNoConstraints(
            worker_id=0, data=data,
            maxeval=maxeval, nsamples=nsamples,
            exp_design=LatinHypercube(dim=data.dim, npts=2*(data.dim+1)),
            response_surface=RSUnitbox(RBFInterpolant(surftype=CubicRBFSurface, maxp=maxeval),data),
            sampling_method=CandidateDYCORS(data=data, numcand=100*data.dim))

    # Launch the threads and give them access to the objective function
    for _ in range(nthreads):
        worker = BasicWorkerThread(controller, data.objfunction)
        controller.launch_worker(worker)

    # Run the optimization strategy
    result = controller.run()

    print('Best value found: {0}'.format(result.value))
    print('Best solution found: {0}\n'.format(
        np.array_str(result.params[0], max_line_width=np.inf,
                     precision=5, suppress_small=True)))

if __name__ == '__main__':
    main()



"""
.. module:: test_problems
  :synopsis: Test problems for multi-modal and
             box-constrained global optimization
.. moduleauthor:: David Eriksson <dme65@cornell.edu>
                 David Bindel <bindel@cornell.edu>

:Module: test_problems
:Author: David Eriksson <dme65@cornell.edu>
    David Bindel <bindel@cornell.edu>
"""

import random
from time import time
import numpy as np
from utils import check_opt_prob

# ========================= 3-dimensional =======================


class Hartman3:
    """Hartman 3 function

    Details: http://www.sfu.ca/~ssurjano/hart3.html

    Global optimum: :math:`f(0.114614,0.555649,0.852547)=-3.86278`

    :ivar dim: Number of dimensions
    :ivar xlow: Lower bound constraints
    :ivar xup: Upper bound constraints
    :ivar info: Problem information:
    :ivar min: Global optimum
    :ivar integer: Integer variables
    :ivar continuous: Continuous variables:
    :ivar constraints: Whether there are non-bound constraints
    """
    def __init__(self, dim=3):
        self.xlow = np.zeros(3)
        self.xup = np.ones(3)
        self.dim = 3
        self.info = "3-dimensional Hartman function \nGlobal optimum: " +\
                    "f(0.114614,0.555649,0.852547) = -3.86278"
        self.min = -3.86278
        self.integer = []
        self.continuous = np.arange(0, 3)
        check_opt_prob(self)

    def objfunction(self, x):
        """Evaluate the Hartman 3 function  at x

        :param x: Data point
        :return: Value at x
        """
        if len(x) != self.dim:
            raise ValueError('Dimension mismatch')
        alpha = np.array([1, 1.2, 3, 3.2])
        A = np.matrix([[3.0, 10.0, 30.0], [0.1, 10.0, 35.0],
                       [3.0, 10.0, 30.0], [0.1, 10.0, 35.0]])
        P = np.matrix([[0.3689, 0.1170, 0.2673],
                       [0.4699, 0.4387, 0.747],
                       [0.1091, 0.8732, 0.5547],
                       [0.0381, 0.5743, 0.8828]])
        outer = 0
        for ii in range(4):
            inner = 0
            for jj in range(3):
                xj = x[jj]
                Aij = A[ii, jj]
                Pij = P[ii, jj]
                inner += Aij * ((xj-Pij) ** 2)
            outer += alpha[ii] * np.exp(-inner)
        return -outer

# ========================= n-dimensional =======================


class Rastrigin:
    """Rastrigin function

    .. math::
        f(x_1,\\ldots,x_n)=10n-\\sum_{i=1}^n (x_i^2 - 10 \\cos(2 \\pi x_i))

    subject to

    .. math::
        -5.12 \\leq x_i \\leq 5.12

    Global optimum: :math:`f(0,0,...,0)=0`

    :ivar dim: Number of dimensions
    :ivar xlow: Lower bound constraints
    :ivar xup: Upper bound constraints
    :ivar info: Problem information:
    :ivar min: Global optimum
    :ivar integer: Integer variables
    :ivar continuous: Continuous variables:
    :ivar constraints: Whether there are non-bound constraints
    """
    def __init__(self, dim=10):
        self.xlow = -5.12 * np.ones(dim)
        self.xup = 5.12 * np.ones(dim)
        self.dim = dim
        self.info = str(dim)+"-dimensional Rastrigin function \n" +\
                             "Global optimum: f(0,0,...,0) = 0"
        self.min = 0
        self.integer = []
        self.continuous = np.arange(0, dim)
        check_opt_prob(self)

    def objfunction(self, x):
        """Evaluate the Rastrigin function  at x

        :param x: Data point
        :return: Value at x
        """
        if len(x) != self.dim:
            raise ValueError('Dimension mismatch')
        return 10 * self.dim + sum(x**2 - 10 * np.cos(2 * np.pi * x))


class Ackley:
    """Ackley function

    .. math::
        f(x_1,\\ldots,x_n) = -20\\exp\\left( -0.2 \\sqrt{\\frac{1}{n} \
        \\sum_{j=1}^n x_j^2} \\right) -\\exp \\left( \\frac{1}{n} \
        \\sum{j=1}^n \\cos(2 \\pi x_j) \\right) + 20 - e

    subject to

    .. math::
        -15 \\leq x_i \\leq 20

    Global optimum: :math:`f(0,0,...,0)=0`

    :ivar dim: Number of dimensions
    :ivar xlow: Lower bound constraints
    :ivar xup: Upper bound constraints
    :ivar info: Problem information:
    :ivar min: Global optimum
    :ivar integer: Integer variables
    :ivar continuous: Continuous variables:
    :ivar constraints: Whether there are non-bound constraints
    """
    def __init__(self, dim=10):
        self.xlow = -15 * np.ones(dim)
        self.xup = 20 * np.ones(dim)
        self.dim = dim
        self.info = str(dim)+"-dimensional Ackley function \n" +\
                             "Global optimum: f(0,0,...,0) = 0"
        self.min = 0
        self.integer = []
        self.continuous = np.arange(0, dim)
        check_opt_prob(self)

    def objfunction(self, x):
        """Evaluate the Ackley function  at x

        :param x: Data point
        :return: Value at x
        """
        if len(x) != self.dim:
            raise ValueError('Dimension mismatch')
        n = float(len(x))
        return -20.0 * np.exp(-0.2*np.sqrt(np.sum(x**2)/n)) - \
            np.exp(np.sum(np.cos(2.0*np.pi*x))/n) + 20 + np.exp(1)


class Michalewicz:
    """Michalewicz function

    """
    def __init__(self, dim=10):
        self.xlow = np.zeros(dim)
        self.xup = np.pi * np.ones(dim)
        self.dim = dim
        self.info = str(dim)+"-dimensional Michalewicz function \n" +\
                             "Global optimum: ??"
        self.min = np.NaN
        self.integer = []
        self.continuous = np.arange(0, dim)
        check_opt_prob(self)

    def objfunction(self, x):
        """Evaluate the Michalewicz function  at x

        :param x: Data point
        :return: Value at x
        """
        if len(x) != self.dim:
            raise ValueError('Dimension mismatch')
        return -np.sum(np.sin(x) * (np.sin(((1+np.arange(self.dim))
                                            * x**2)/np.pi)) ** 20)


class Levy:
    def __init__(self, dim=10):
        self.xlow = -5 * np.ones(dim)
        self.xup = 5 * np.ones(dim)
        self.dim = dim
        self.info = str(dim)+"-dimensional Levy function \n" +\
                             "Global optimum: ?"
        self.integer = []
        self.continuous = np.arange(0, dim)
        check_opt_prob(self)

    def objfunction(self, x):
        """Evaluate the Levy function  at x

        :param x: Data point
        :return: Value at x
        """
        if len(x) != self.dim:
            raise ValueError('Dimension mismatch')
        a = 20
        b = 0.2
        d = 50
        return -a * np.exp(-b * np.sqrt(np.sum(x ** 2)/d)) -\
            np.exp(np.sum(np.cos(2 * np.pi * x))/d)


class Griewank:
    """Griewank function

    .. math::
        f(x_1,\\ldots,x_n) = 1 + \\frac{1}{4000} \\sum_{j=1}^n x_j^2 - \
        \\prod_{j=1}^n \\cos \\left( \\frac{x_i}{\\sqrt{i}} \\right)

    subject to

    .. math::
        -512 \\leq x_i \\leq 512

    Global optimum: :math:`f(0,0,...,0)=0`

    :ivar dim: Number of dimensions
    :ivar xlow: Lower bound constraints
    :ivar xup: Upper bound constraints
    :ivar info: Problem information:
    :ivar min: Global optimum
    :ivar integer: Integer variables
    :ivar continuous: Continuous variables:
    :ivar constraints: Whether there are non-bound constraints
    """
    def __init__(self, dim=10):
        self.xlow = -512 * np.ones(dim)
        self.xup = 512 * np.ones(dim)
        self.dim = dim
        self.info = str(dim)+"-dimensional Griewank function \n" +\
                             "Global optimum: f(0,0,...,0) = 0"
        self.min = 0
        self.integer = []
        self.continuous = np.arange(0, dim)
        check_opt_prob(self)

    def objfunction(self, x):
        """Evaluate the Griewank function  at x

        :param x: Data point
        :return: Value at x
        """
        if len(x) != self.dim:
            raise ValueError('Dimension mismatch')
        total = 1
        for i, y in enumerate(x):
            total *= np.cos(y / np.sqrt(i+1))
        return 1.0 / 4000.0 * sum([y**2 for y in x]) - total + 1


class Rosenbrock:
    """Rosenbrock function

    .. math::
        f(x_1,\\ldots,x_n) = \\sum_{j=1}^{n-1} \
        \\left( 100(x_j^2-x_{j+1})^2 + (1-x_j)^2 \\right)

    subject to

    .. math::
        -2.048 \\leq x_i \\leq 2.048

    Global optimum: :math:`f(1,1,...,1)=0`

    :ivar dim: Number of dimensions
    :ivar xlow: Lower bound constraints
    :ivar xup: Upper bound constraints
    :ivar info: Problem information:
    :ivar min: Global optimum
    :ivar integer: Integer variables
    :ivar continuous: Continuous variables:
    :ivar constraints: Whether there are non-bound constraints
    """
    def __init__(self, dim=10):
        self.xlow = -2.048 * np.ones(dim)
        self.xup = 2.048 * np.ones(dim)
        self.dim = dim
        self.info = str(dim)+"-dimensional Rosenbrock function \n" +\
                             "Global optimum: f(1,1,...,1) = 0"
        self.min = 0
        self.integer = []
        self.continuous = np.arange(0, dim)
        check_opt_prob(self)

    def objfunction(self, x):
        """Evaluate the Rosenbrock function  at x

        :param x: Data point
        :return: Value at x
        """
        if len(x) != self.dim:
            raise ValueError('Dimension mismatch')
        total = 0
        for i in range(len(x) - 1):
            total += 100 * (x[i] ** 2 - x[i+1]) ** 2 + (x[i] - 1) ** 2
        return total


class Schwefel:
    """Schwefel function

    .. math::
        f(x_1,\\ldots,x_n) = \\sum_{j=1}^{n} \
        \\left( -x_j \\sin(\\sqrt{|x_j|}) \\right) + 418.982997 n

    subject to

    .. math::
        -512 \\leq x_i \\leq 512

    Global optimum: :math:`f(420.968746,420.968746,...,420.968746)=0`

    :ivar dim: Number of dimensions
    :ivar xlow: Lower bound constraints
    :ivar xup: Upper bound constraints
    :ivar info: Problem information:
    :ivar min: Global optimum
    :ivar integer: Integer variables
    :ivar continuous: Continuous variables:
    :ivar constraints: Whether there are non-bound constraints
    """
    def __init__(self, dim=10):
        self.xlow = -512 * np.ones(dim)
        self.xup = 512 * np.ones(dim)
        self.dim = dim
        self.info = str(dim)+"-dimensional Schwefel function \n" +\
                             "Global optimum: f(420.968746,...,420.968746) = 0"
        self.min = 0
        self.integer = []
        self.continuous = np.arange(0, dim)
        check_opt_prob(self)

    def objfunction(self, x):
        """Evaluate the Schwefel function  at x

        :param x: Data point
        :return: Value at x
        """
        if len(x) != self.dim:
            raise ValueError('Dimension mismatch')
        return 418.9829 * self.dim - \
            sum([y * np.sin(np.sqrt(abs(y))) for y in x])


class Sphere:
    """Sphere function

    .. math::
        f(x_1,\\ldots,x_n) = \\sum_{j=1}^n x_j^2

    subject to

    .. math::
        -5.12 \\leq x_i \\leq 5.12

    Global optimum: :math:`f(0,0,...,0)=0`

    :ivar dim: Number of dimensions
    :ivar xlow: Lower bound constraints
    :ivar xup: Upper bound constraints
    :ivar info: Problem information:
    :ivar min: Global optimum
    :ivar integer: Integer variables
    :ivar continuous: Continuous variables:
    :ivar constraints: Whether there are non-bound constraints
    """
    def __init__(self, dim=10):
        self.xlow = -5.12 * np.ones(dim)
        self.xup = 5.12 * np.ones(dim)
        self.dim = dim
        self.info = str(dim)+"-dimensional Sphere function \n" +\
                             "Global optimum: f(0,0,...,0) = 0"
        self.min = 0
        self.integer = []
        self.continuous = np.arange(0, dim)
        check_opt_prob(self)

    def objfunction(self, x):
        """Evaluate the Sphere function  at x

        :param x: Data point
        :return: Value at x
        """
        if len(x) != self.dim:
            raise ValueError('Dimension mismatch')
        return np.sum(x ** 2)


class Exponential:
    """Exponential function

    .. math::
        f(x_1,\\ldots,x_n) = \\sum_{j=1}^n e^{jx_j} - \\sum_{j=1} e^{-5.12 j}

    subject to

    .. math::
        -5.12 \\leq x_i \\leq 5.12

    Global optimum: :math:`f(0,0,...,0)=0`

    :ivar dim: Number of dimensions
    :ivar xlow: Lower bound constraints
    :ivar xup: Upper bound constraints
    :ivar info: Problem information:
    :ivar min: Global optimum
    :ivar integer: Integer variables
    :ivar continuous: Continuous variables:
    :ivar constraints: Whether there are non-bound constraints
    """
    def __init__(self, dim=10):
        self.xlow = -5.12 * np.ones(dim)
        self.xup = 5.12 * np.ones(dim)
        self.dim = dim
        self.info = str(dim)+"-dimensional Exponential function \n" +\
                             "Global optimum: f(-5.12,-5.12,...,-5.12) = 0"
        self.min = 0
        self.integer = []
        self.continuous = np.arange(0, dim)
        check_opt_prob(self)

    def objfunction(self, x):
        """Evaluate the Exponential function  at x

        :param x: Data point
        :return: Value at x
        """
        if len(x) != self.dim:
            raise ValueError('Dimension mismatch')
        total = 0
        for i in range(len(x)):
            total += np.exp((i+1)*x[i-1]) - np.exp(-5.12*(i+1))
        return total


class StyblinskiTang:
    """StyblinskiTang function

    .. math::
        f(x_1,\\ldots,x_n) = \\frac{1}{2} \\sum_{j=1}^n  \
        \\left(x_j^4 -16x_j^2 +5x_j \\right)

    subject to

    .. math::
        -5 \\leq x_i \\leq 5

    Global optimum: :math:`f(-2.903534,-2.903534,...,-2.903534)=\
    -39.16599 \\cdot n`

    :ivar dim: Number of dimensions
    :ivar xlow: Lower bound constraints
    :ivar xup: Upper bound constraints
    :ivar info: Problem information:
    :ivar min: Global optimum
    :ivar integer: Integer variables
    :ivar continuous: Continuous variables:
    :ivar constraints: Whether there are non-bound constraints
    """
    def __init__(self, dim=10):
        self.xlow = -5 * np.ones(dim)
        self.xup = 5 * np.ones(dim)
        self.dim = dim
        self.info = str(dim)+"-dimensional Styblinski-Tang function \n" +\
                             "Global optimum: f(-2.903534,...,-2.903534) = " +\
                             str(-39.16599*dim)
        self.min = -39.16599*dim
        self.integer = []
        self.continuous = np.arange(0, dim)
        check_opt_prob(self)

    def objfunction(self, x):
        """Evaluate the StyblinskiTang function  at x

        :param x: Data point
        :return: Value at x
        """
        if len(x) != self.dim:
            raise ValueError('Dimension mismatch')
        return 0.5*np.sum(x ** 4 - 16 * x ** 2 + 5 * x)


class Quartic:
    """Quartic function

    .. math::
        f(x_1,\\ldots,x_n) = \\sum_{j=1}^n j x_j^4 + random[0,1)

    subject to

    .. math::
        -1.28 \\leq x_i \\leq 1.28

    Global optimum: :math:`f(0,0,...,0)=0+noise`

    :ivar dim: Number of dimensions
    :ivar xlow: Lower bound constraints
    :ivar xup: Upper bound constraints
    :ivar info: Problem information:
    :ivar min: Global optimum
    :ivar integer: Integer variables
    :ivar continuous: Continuous variables:
    :ivar constraints: Whether there are non-bound constraints
    """
    def __init__(self, dim=10):
        self.xlow = -1.28 * np.ones(dim)
        self.xup = 1.28 * np.ones(dim)
        self.dim = dim
        self.info = str(dim)+"-dimensional Quartic function \n" +\
                             "Global optimum: f(0,0,...,0) = 0+noise"
        self.min = 0
        self.integer = []
        self.continuous = np.arange(0, dim)
        self.prng = random.Random()
        self.prng.seed(time())
        check_opt_prob(self)

    def objfunction(self, x):
        """Evaluate the Quartic function  at x

        :param x: Data point
        :return: Value at x
        """
        if len(x) != self.dim:
            raise ValueError('Dimension mismatch')
        total = 0.0
        for i in xrange(len(x)):
            total += (i+1.0) * x[i]**4.0
        return total + self.prng.uniform(0, 1)


class Whitley:
    """Quartic function

    .. math::
        f(x_1,\\ldots,x_n) = \\sum_{i=1}^n \\sum_{j=1}^n \
        \\left( \\frac{(100(x_i^2-x_j)^2+(1-x_j)^2)^2}{4000} \
        - \\cos(100(x_i^2-x_j)^2 + (1-x_j)^2 ) + 1 \\right)

    subject to

    .. math::
        -10.24 \\leq x_i \\leq 10.24

    Global optimum: :math:`f(1,1,...,1)=0`

    :ivar dim: Number of dimensions
    :ivar xlow: Lower bound constraints
    :ivar xup: Upper bound constraints
    :ivar info: Problem information:
    :ivar min: Global optimum
    :ivar integer: Integer variables
    :ivar continuous: Continuous variables:
    :ivar constraints: Whether there are non-bound constraints
    """
    def __init__(self, dim=10):
        self.xlow = -10.24 * np.ones(dim)
        self.xup = 10.24 * np.ones(dim)
        self.dim = dim
        self.info = str(dim)+"-dimensional Whitley function \n" +\
                             "Global optimum: f(1,1,...,1) = 0"
        self.min = 0
        self.integer = []
        self.continuous = np.arange(0, dim)
        check_opt_prob(self)

    def objfunction(self, x):
        """Evaluate the Whitley function  at x

        :param x: Data point
        :return: Value at x
        """
        if len(x) != self.dim:
            raise ValueError('Dimension mismatch')
        total = 0
        for i in range(len(x)):
            for j in range(len(x)):
                temp = 100*((x[i]**2)-x[j]) + (1-x[j])**2
                total += (float(temp**2)/4000.0) - np.cos(temp) + 1
        return total


class SchafferF7:
    """SchafferF7 function

    .. math::
        f(x_1,\\ldots,x_n) = \\left[\\frac{1}{n-1}\\sqrt{s_i} \
        \\cdot (\\sin(50.0s_i^{\\frac{1}{5}})+1)\\right]^2

    where

    .. math::
        s_i = \\sqrt{x_i^2 + x_{i+1}^2}

    subject to

    .. math::
        -100 \\leq x_i \\leq 100

    Global optimum: :math:`f(0,0,...,0)=0`

    :ivar dim: Number of dimensions
    :ivar xlow: Lower bound constraints
    :ivar xup: Upper bound constraints
    :ivar info: Problem information:
    :ivar min: Global optimum
    :ivar integer: Integer variables
    :ivar continuous: Continuous variables:
    :ivar constraints: Whether there are non-bound constraints
    """
    def __init__(self, dim=10):
        self.xlow = -100 * np.ones(dim)
        self.xup = 100 * np.ones(dim)
        self.dim = dim
        self.info = str(dim)+"-dimensional SchafferF7 function \n" +\
                             "Global optimum: f(0,0,...,0) = 0"
        self.min = 0
        self.integer = []
        self.continuous = np.arange(0, dim)
        check_opt_prob(self)

    def objfunction(self, x):
        """Evaluate the SchafferF7 function  at x

        :param x: Data point
        :return: Value at x
        """
        if len(x) != self.dim:
            raise ValueError('Dimension mismatch')
        total = 0
        normalizer = 1.0/float(len(x)-1)
        for i in range(len(x)-1):
            si = np.sqrt(x[i]**2 + x[i+1]**2)
            total += (normalizer * np.sqrt(si) * (np.sin(50*si**0.20) + 1))**2
        return total

# ============================== Constraints ==============================


class Keane:
    """Keane's "bump" function

    .. math::
        f(x_1,\\ldots,x_n) = -\\left| \\frac{\\sum_{j=1}^n \\cos^4(x_j) - \
        2 \\prod_{j=1}^n \\cos^2(x_j)}{\\sqrt{\\sum_{j=1}^n jx_j^2}} \\right|

    subject to

    .. math::
        -100 \\leq x_i \\leq 100

    .. math::
        0.75 - \\prod_{j=1}^n x_j < 0

    .. math::
        \\sum_{j=1}^n x_j - 7.5n < 0


    Global optimum: -0.835 for large n

    :ivar dim: Number of dimensions
    :ivar xlow: Lower bound constraints
    :ivar xup: Upper bound constraints
    :ivar info: Problem information:
    :ivar min: Global optimum
    :ivar integer: Integer variables
    :ivar continuous: Continuous variables:
    :ivar constraints: Whether there are non-bound constraints
    """
    def __init__(self, dim=10):
        self.xlow = np.zeros(dim)
        self.xup = 5 * np.ones(dim)
        self.dim = dim
        self.min = -0.835
        self.integer = []
        self.continuous = np.arange(0, dim)
        self.info = str(dim)+"-dimensional Keane bump function \n" +\
                             "Global optimum: -0.835 for large n"
        check_opt_prob(self)

    # Compute the value of the constraint functions at the given point
    # Returns an array of size npts x nconstraints
    def eval_ineq_constraints(self, x):
        vec = np.zeros((x.shape[0], 2))
        vec[:, 0] = 0.75 - np.prod(x)
        vec[:, 1] = np.sum(x) - 7.5 * self.dim
        return vec

    # Compute the derivative of the constraint functions
    # Returns an array of size npts x nconstraints x ndims
    def deriv_ineq_constraints(self, x):
        vec = np.zeros((x.shape[0], 2, x.shape[1]))
        for i in range(x.shape[0]):
            xx = x[i, :]
            for j in range(x.shape[1]):
                vec[i, 0, j] = -np.prod(np.hstack((xx[:j], xx[j+1:])))
                vec[i, 1, j] = 1.0
        return vec

    # Evaluate the objective function for a single data point
    def objfunction(self, x):
        """Evaluate the Keane function  at x

        :param x: Data point
        :return: Value at x
        """
        if len(x) != self.dim:
                raise ValueError('Dimension mismatch')
        n = len(x)
        return -abs((sum(np.cos(x)**4)-2 * np.prod(np.cos(x)**2)) /
                    max([1E-10, np.sqrt(np.dot(1+np.arange(n), x**2))]))


class LinearMI:
    """This is a linear mixed integer problem with non-bound constraints

    There are 5 variables, the first 3 are discrete and the last 2
    are continuous.

    Global optimum: :math:`f(1,0,0,0,0) = -1`
    """
    def __init__(self):
        self.xlow = np.zeros(5)
        self.xup = np.array([10, 10, 10, 1, 1])
        self.dim = 5
        self.min = -1
        self.integer = np.arange(0, 3)
        self.continuous = np.arange(3, 5)
        self.info = str(self.dim)+"-dimensional Linear MI \n" +\
                                  "Global optimum: f(1,0,0,0,0) = -1\n" +\
                                  str(len(self.integer)) + " integer variables"
        check_opt_prob(self)

    def eval_ineq_constraints(self, x):
        vec = np.zeros((x.shape[0], 3))
        vec[:, 0] = x[:, 0] + x[:, 2] - 1.6
        vec[:, 1] = 1.333 * x[:, 1] + x[:, 3] - 3
        vec[:, 2] = - x[:, 2] - x[:, 3] + x[:, 4]
        return vec

    def objfunction(self, x):
        """Evaluate the LinearMI function  at x

        :param x: Data point
        :return: Value at x
        """
        if len(x) != self.dim:
            raise ValueError('Dimension mismatch')
        return - x[0] + 3 * x[1] + 1.5 * x[2] + 2 * x[3] - 0.5 * x[4]


if __name__ == "__main__":
    print("\n========================= Hartman3 =======================")
    fun = Hartman3()
    print(fun.info)
    print("Hartman3(1,1,1) = " + str(fun.objfunction(np.array([1, 1, 1]))))
    print("Continuous variables: " + str(fun.continuous))
    print("Integer variables: " + str(fun.integer))

    print("\n========================= Rastrigin =======================")
    fun = Rastrigin(dim=3)
    print(fun.info)
    print("Rastrigin(1,1,1) = " + str(fun.objfunction(np.array([1, 1, 1]))))
    print("Continuous variables: " + str(fun.continuous))
    print("Integer variables: " + str(fun.integer))

    print("\n========================= Ackley =======================")
    fun = Ackley(dim=3)
    print(fun.info)
    print("Ackley(1,1,1) = " + str(fun.objfunction(np.array([1, 1, 1]))))

    print("\n========================= Griewank =======================")
    fun = Griewank(dim=3)
    print(fun.info)
    print("Griewank(1,1,1) = " + str(fun.objfunction(np.array([1, 1, 1]))))
    print("Continuous variables: " + str(fun.continuous))
    print("Integer variables: " + str(fun.integer))

    print("\n========================= Rosenbrock =======================")
    fun = Rosenbrock(dim=3)
    print(fun.info)
    print("Rosenbrock(1,1,1) = " + str(fun.objfunction(np.array([1, 1, 1]))))
    print("Continuous variables: " + str(fun.continuous))
    print("Integer variables: " + str(fun.integer))

    print("\n========================= Schwefel =======================")
    fun = Schwefel(dim=3)
    print(fun.info)
    print("Schwefel(1,1,1) = " + str(fun.objfunction(np.array([1, 1, 1]))))
    print("Continuous variables: " + str(fun.continuous))
    print("Integer variables: " + str(fun.integer))

    print("\n========================= Sphere =======================")
    fun = Sphere(dim=3)
    print(fun.info)
    print("Sphere(1,1,1) = " + str(fun.objfunction(np.array([1, 1, 1]))))
    print("Continuous variables: " + str(fun.continuous))
    print("Integer variables: " + str(fun.integer))

    print("\n========================= Exponential =======================")
    fun = Exponential(dim=3)
    print(fun.info)
    print("Exponential(1,1,1) = " + str(fun.objfunction(np.array([1, 1, 1]))))
    print("Continuous variables: " + str(fun.continuous))
    print("Integer variables: " + str(fun.integer))

    print("\n======================= Styblinski-Tang =====================")
    fun = StyblinskiTang(dim=3)
    print(fun.info)
    print("StyblinskiTang(1,1,1) = " +
          str(fun.objfunction(np.array([1, 1, 1]))))
    print("Continuous variables: " + str(fun.continuous))
    print("Integer variables: " + str(fun.integer))

    print("\n========================= Whitley =======================")
    fun = Whitley(dim=3)
    print(fun.info)
    print("Whitley(1,1,1) = " + str(fun.objfunction(np.array([1, 1, 1]))))
    print("Continuous variables: " + str(fun.continuous))
    print("Integer variables: " + str(fun.integer))

    print("\n========================= Quartic =======================")
    fun = Quartic(dim=3)
    print(fun.info)
    print("Quartic(1,1,1) = " + str(fun.objfunction(np.array([1, 1, 1]))))
    print("Continuous variables: " + str(fun.continuous))
    print("Integer variables: " + str(fun.integer))

    print("\n========================= SchafferF7 =======================")
    fun = SchafferF7(dim=3)
    print(fun.info)
    print("SchafferF7(1,1,1) = " + str(fun.objfunction(np.array([1, 1, 1]))))
    print("Continuous variables: " + str(fun.continuous))
    print("Integer variables: " + str(fun.integer))

    print("\n========================= Liner MI =======================")
    fun = LinearMI()
    print(fun.info)
    print("Linear_MI(1,1,1,1,1) = " +
          str(fun.objfunction(np.array([1, 1, 1, 1, 1]))))
    print("Continuous variables: " + str(fun.continuous))
    print("Integer variables: " + str(fun.integer))

    print("\n========================= Michalewicz =======================")
    fun = Michalewicz(dim=2)
    print(fun.info)
    print("Michalewicz(2.20, 1.57) = " +
          str(fun.objfunction(np.array([2.20, 1.57]))))
    print("Continuous variables: " + str(fun.continuous))
    print("Integer variables: " + str(fun.integer))



"""
.. module:: test_matlab_engine
  :synopsis: Test with MATLAB objective function
.. moduleauthor:: David Eriksson <dme65@cornell.edu>
"""

from pySOT import *
from poap.controller import ThreadController, ProcessWorkerThread
import numpy as np
import os.path
import matlab_wrapper


class MatlabWorker(ProcessWorkerThread):
    def handle_eval(self, record):
        try:
            self.matlab.put('x', record.params[0])
            self.matlab.eval('matlab_ackley')
            val = self.matlab.get('val')
            if np.isnan(val):
                raise ValueError()
            self.finish_success(record, val)
        except:
            self.finish_cancelled(record)


def main():
    if not os.path.exists("./logfiles"):
        os.makedirs("logfiles")
    if os.path.exists("./logfiles/test_matlab_engine.log"):
        os.remove("./logfiles/test_matlab_engine.log")
    logging.basicConfig(filename="./logfiles/test_matlab_engine.log",
                        level=logging.INFO)

    print("\nNumber of threads: 4")
    print("Maximum number of evaluations: 500")
    print("Sampling method: CandidateDYCORS")
    print("Experimental design: Latin Hypercube")
    print("Surrogate: Cubic RBF, domain scaled to unit box")

    nthreads = 4
    maxeval = 500

    data = Ackley(dim=10)
    print(data.info)

    # Use the serial controller (uses only one thread)
    controller = ThreadController()
    controller.strategy = \
        SyncStrategyNoConstraints(
            worker_id=0, data=data,
            maxeval=maxeval, nsamples=nthreads,
            exp_design=LatinHypercube(dim=data.dim, npts=2*(data.dim+1)),
            response_surface=RSUnitbox(RBFInterpolant(surftype=CubicRBFSurface, maxp=maxeval),data),
            sampling_method=CandidateDYCORS(data=data, numcand=100*data.dim))

    print("\nNOTE: You may need to specify the matlab_root keyword in "
          "order \n      to start a MATLAB session using the matlab_wrapper "
          "module\n")

    # We need to tell MATLAB where the script is
    mfile_location = os.getcwd()

    # Launch the threads
    for _ in range(nthreads):
        worker = MatlabWorker(controller)
        worker.matlab = matlab_wrapper.MatlabSession(options='-nojvm')
        worker.matlab.workspace.addpath(mfile_location)
        controller.launch_worker(worker)

    # Run the optimization strategy
    result = controller.run()

    # Print the final result
    print('Best value found: {0}'.format(result.value))
    print('Best solution found: {0}'.format(
        np.array_str(result.params[0], max_line_width=np.inf,
                     precision=5, suppress_small=True)))


if __name__ == '__main__':
    main()



"""
.. module:: rbf_surface
   :synopsis: RBF interpolation surface base classes
.. moduleauthor:: David Bindel <bindel@cornell.edu>

:Module: rbf_surface
:Author: David Bindel <bindel@cornell.edu>
"""


import numpy as np
import scipy.linalg as la
import scipy.spatial.distance as dist
from array_manager import ArrayManager


class BaseRBFSystem(object):
    """Base class for RBF system

    The RBF system is the set of centers and basis functions
    (polynomial and RBF), and the associated linear system.
    It does not involve actual function values or coefficients
    for interpolants.  One RBF system object may be shared among
    multiple interpolants.

    Derived classes should overload:
    - fill_basis - evaluate basis at points
    - basis_dot - compute linear combination of basis values
    - dbasis_dot - compute linear combination of basis gradients

    Derived classes may also overload the factor and solve routines.

    Attributes:
        new_points: Number of points since the last update
    """

    def __init__(self, dim, mtail, maxpts):
        """Initialize the system

        Args:
            dim: Dimension of the system
            mtail: Dimension of the polynomial tail
            maxpts: Initial estimate of maximum number of centers
        """
        self._mtail = mtail
        self._xmgr = ArrayManager((0, dim), (maxpts, dim))
        self._Mmgr = ArrayManager((self.nsys, self.nsys))
        self.new_points = 0

    @property
    def x(self):
        "Array of points"
        return self._xmgr.view

    @property
    def M(self):
        "System matrix"
        return self._Mmgr.view

    @property
    def dim(self):
        "Dimension of the ambient space"
        return self._xmgr.shape[1]

    @property
    def npt(self):
        "Number of centers"
        return self._xmgr.shape[0]

    @property
    def mtail(self):
        "Dimension of the polynomial tail space"
        return self._mtail

    @property
    def nsys(self):
        "System size"
        return self.npt + self.mtail

    @property
    def dirty(self):
        "Return true if there are new points"
        return self.new_points > 0

    def add_centers(self, x):
        """Add centers to the RBF interpolant

        Args:
            x: One or more new center locations (one per row if multiple)
        """
        if len(x.shape) == 1:
            self._xmgr.append(np.array([x]))
            self.new_points += 1
        else:
            self._xmgr.append(x)
            self.new_points += x.shape[0]

    def dist(self, y):
        """Compute distances of all points to y

        Args:
            y: One or more points (one per row if multiple)
        """
        if len(y.shape) == 1:
            return dist.cdist(self.x, np.expand_dims(y, axis=0)).ravel()
        else:
            return dist.cdist(self.x, y)

    def basis(self, y, d=None):
        """Compute the vector of basis functions at y

        Args:
            y: Point at which to evaluate basis functions
            d: Distance from y to RBF (not required)
        """
        b = np.zeros((self.nsys, 1))
        self.fill_basis(b, np.expand_dims(y, axis=0), d=None)
        b.shape = (self.nsys,)
        return b

    def compliance(self, y):
        """Get the compliance of the surface at y

        Args:
            y: Point at which to compute surface compliance
        """
        self.update()
        b = self.basis(y)
        return -np.dot(b, self.solve(b))

    def update(self):
        """Update the system matrix if points have been added
        """
        if self.new_points == 0:
            return
        x, nnew, Mmgr = self.x, self.new_points, self._Mmgr
        Mmgr.resize((self.nsys, self.nsys))
        M = Mmgr.view
        self.fill_basis(M[:, -nnew:], x[-nnew:, :])
        M[-nnew:, :] = M[:, -nnew:].T
        self.refactor()
        self.new_points = 0

    def refactor(self):
        """Compute factorization
        """
        eta = min(1e-5, 1e-16 * np.sqrt(la.norm(self.M, 1) * la.norm(self.M, np.inf))) * np.eye(self.M.shape[0])
        self.lupiv = la.lu_factor(self.M + eta)

    def solve(self, rhs):
        """Solve with the system matrix

        Args:
            rhs: Right hand side in system

        Returns:
            solution to M*x = rhs
        """
        self.update()
        return la.lu_solve(self.lupiv, rhs)

    def fill_basis(self, B, y, d=None):
        """Compute a matrix of basis functions at y

        Args:
            B: Matrix to fill with basis function values
            y: Points at which to evaluate basis functions
            d: Distance from y to centers (optional)
        """
        pass

    def basis_dot(self, v, y, d=None):
        """Form a linear combination of basis functions

        Args:
            v: Coefficient vector in linear combination
            y: Point at which to evaluate basis functions
            d: Distance from y to centers (optional)
        """
        pass

    def dbasis_dot(self, v, y, d=None):
        """Form a linear combination of the basis gradients at y

        Args:
            v: Coefficient vector in linear combination
            y: Point at which to evaluate basis function gradients
            d: Distance from y to centers (optional)
        """
        pass


class SimpleRBFSystem(BaseRBFSystem):
    """Base class for RBF systems with simple polynomial tails

    Assumes that we use a low-degree polynomial tail (at most linear)
    with a standard basis.

    Derived classes should overload:
    - phi - evaluate RBF function
    - dphi_div_r - evaluate phi'(r)/r

    Attributes:
       dpoly: Degree bound for polynomial tail (derived class variable)
    """

    def __init__(self, dim, maxpts):
        """Initialize the system

        Args:
            dim: Dimension of the system
            dpoly: One more than the max poly degree (e.g. 2 for linear)
            maxpts: Initial estimate of maximum number of centers
        """
        dpoly = self.dpoly
        if dpoly == 0:
            mtail = 0
        elif dpoly == 1:
            mtail = 1
        elif dpoly == 2:
            mtail = dim + 1
        super(SimpleRBFSystem, self).__init__(dim, mtail, maxpts)

    def fill_basis(self, B, y, d=None):
        """Compute a matrix of basis functions at y

        Args:
            B: Matrix to fill with basis function values
            y: Points at which to evaluate basis functions
            d: Distance from y to centers (optional)
        """
        m = self.mtail
        if self.dpoly > 0:
            B[0, :] = 1
        if self.dpoly > 1:
            B[1:m, :] = y.T
        B[m:, :] = d if d is not None else self.dist(y)
        B[m:, :] = self.phi(B[m:, :])
        return B

    def basis_dot(self, v, y, d=None):
        """Form a linear combination of basis functions

        Args:
            v: Coefficient vector in linear combination
            y: Point at which to evaluate basis functions
            d: Distance from y to centers (optional)
        """

        if d is None:
            d = self.dist(y)
        m = self.mtail
        sy = np.dot(v[m:], self.phi(d))
        if self.dpoly > 0:
            sy += v[0]
        if self.dpoly > 1:
            sy += np.dot(y, v[1:m])
        return sy

    def dbasis_dot(self, v, y, d=None):
        """Form a linear combination of basis function gradients

        Args:
            v: Coefficient vector in linear combination
            y: Point at which to evaluate basis function gradients
            d: Distance from y to centers (optional)
        """
        if d is None:
            d = self.dist(y)
        x, m = self.x, self.mtail
        dsy = np.dot(self.dphi_div_r(d) * v[m:], y-x)
        if self.dpoly > 1:
            dsy += v[1:m]
        return dsy


class RBFSurface(object):
    """RBF surface

    An RBF surface builds an interpolant on top of an RBF system object.

    Attributes:
        rbfs: RBF system object
        coeff: interpolant coefficient vector
        dirty: flag whether there have been changes since coeff computation
    """

    def __init__(self, rbfs):
        """Initialize the surface

        Args:
            rbfs: RBF system object
        """
        self._fxmgr = ArrayManager((rbfs.nsys,))
        self.coeff = None
        self.rbfs = rbfs
        self.dirty = True

    @property
    def x(self):
        "Array of points"
        return self.rbfs.x

    @property
    def fx(self):
        "Array of function values"
        return self._fxmgr.view

    @property
    def dim(self):
        "Dimension of the ambient space"
        return self.rbfs.dim

    @property
    def npt(self):
        "Number of centers"
        return self.rbfs.npt

    def set_fx(self, fx):
        "Replace f with transformed function values"
        self.fx[self.rbfs.mtail:] = fx
        self.dirty = True

    def fit(self):
        "Compute the fit"
        self.coeff = self.rbfs.solve(self.fx)
        self.dirty = False

    def update(self):
        "Update the fit if stale"
        if self.dirty:
            self.fit()

    def add_values(self, fx):
        """Add values to the system

        Args:
            fx: vector of values to append
        """
        self._fxmgr.append(fx)
        self.dirty = True

    def add_centers(self, x):
        """Add a center to the system

        Args:
            x: center to add
        """
        self.rbfs.add_centers(x)

    def add_points(self, x, fx):
        """Add points to the system

        NB: This function is only really safe if the RBF system is not
        shared.  To work with multiple surfaces associated with the same
        RBF system, add centers to the system object and values to the
        surfaces separately.

        Args:
            x: points to add to the RBF system
            fx: vector of values to append
        """
        self.add_centers(x)
        self.add_values(fx)

    def eval(self, y, d=None):
        """Evaluate the surface at y

        Args:
            y: Point or points at which to evaluate

        Returns:
            value of the surface at y
        """
        self.update()
        return self.rbfs.basis_dot(self.coeff, y, d)

    def deriv(self, y, d=None):
        """Evaluate the derivative at y

        Args:
            y: Point or points at which to evaluate

        Returns:
            gradient of the surface at y
        """
        self.update()
        return self.rbfs.dbasis_dot(self.coeff, y, d)

    def seminorm(self):
        "Get the seminorm of the surface"
        self.update()
        c, M = self.coeff, self.rbfs.M
        return np.dot(c, np.dot(M, c))

    def dseminorm(self, y, tau):
        """Compute change in seminorm associated with adding center at y

        Gutmann (2001) suggests choosing sample points in a global
        optimization by finding where the smallest seminorm change
        could be applied to the surface in order to achieve a given
        target value.

        Args:
            y: Point at which to add a hypothetical new center
            tau: Value of s(y) if augmented with hypothetical center

        Returns:
            g: Change in squared seminorm between original and updated surface
            dg: Gradient of g with respect to y

        """
        self.update()
        rbfs = self.rbfs
        d = rbfs.dist(y)
        b = rbfs.basis(y, d)
        w = rbfs.solve(b)
        mu = -np.dot(w, b)
        sy = np.dot(self.coeff, b)
        eta = (sy-tau)/mu
        g = (sy-tau)*eta
        dg = (2*eta)*rbfs.dbasis_dot(self.coeff + eta*w, y, d)
        return g, dg

    def diseminorm(self, y, tau):
        """Compute 1/g(y) where g(y) is computed by dseminorm

        The function h(y) = -1/g(y) is potentially nicer to work
        with than just g(y), as it doesn't have any poles.

        Args:
            y: Point at which to add a hypothetical new center
            tau: Value of s(y) if augmented with hypothetical center

        Returns:
            h: Value of h(y) = 1/g(y)
            dh: Gradient of h with respect to y

        """
        self.update()
        rbfs = self.rbfs
        d = rbfs.dist(y)
        b = rbfs.basis(y, d)
        w = rbfs.solve(b)
        mu = -np.dot(w, b)
        sy = np.dot(self.coeff, b)
        gap = sy-tau
        h = mu/gap**2
        dh = -2/gap**2 * rbfs.dbasis_dot(mu/gap * self.coeff + w, y, d)
        return h, dh


class SimpleRBFSurface(RBFSurface):
    """Simple RBF surface with a private RBF system object.

    Class variables:
        RBFSystem: Name of the RBF system object class

    Attributes:
        rbfs: RBF system object
        coeff: interpolant coefficient vector
        dirty: flag whether there have been changes since coeff computation
    """

    def __init__(self, x=None, fx=None, dim=None, maxpts=100):
        """Initialize the surface.

        Args:
            x: Initial points
            fx: Function values at f
            dim: Dimension of the space (required if x not given)
            maxpts: Estimate of max points (default: 100)
        """
        if x is not None:
            assert dim is None or dim == x.shape[1], "Dimension mismatch"
            assert fx is not None, "Must supply function values"
            dim = x.shape[1]
        rbfs = self.RBFSystem(dim, maxpts)
        super(SimpleRBFSurface, self).__init__(rbfs)
        if x is not None:
            self.add_points(x, fx)



try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

from experimental_design import *
from rbf_interpolant import *
from rbf_surfaces import *
from rs_capped import RSCapped, RSUnitbox
from sampling_methods import *
from test_problems import *
from sot_sync_strategies import *
from ensemble_surrogate import EnsembleSurrogate
from poly_regression import *
from merit_functions import *

try:
    from GUI import GUI
except ImportError:
    pass

try:
    from mars_interpolant import MARSInterpolant
except ImportError:
    pass

try:
    from kriging_interpolant import KrigingInterpolant
except ImportError:
    pass



"""
.. module:: heurisitc_algorithms
   :synopsis: A mixed-integer GA implementation for bound constrained problems
.. moduleauthor:: David Eriksson <dme65@cornell.edu>
"""

from experimental_design import LatinHypercube, SymmetricLatinHypercube
import numpy as np


class GeneticAlgorithm:
    def __init__(self, function, dim, xlow, xup, intvar=None, popsize=100, ngen=100, start="SLHD", proj_fun=None):
        self.nvariables = dim
        self.nindividuals = popsize + (popsize % 2)  # Make sure this is even
        self.lower_boundary = np.array(xlow)
        self.upper_boundary = np.array(xup)
        self.integer_variables = []
        if intvar is not None:
            self.integer_variables = np.array(intvar)
        self.start = start
        self.sigma = 0.2
        self.p_mutation = 1.0/dim
        self.tournament_size = 5
        self.p_cross = 0.9
        self.ngenerations = ngen
        self.function = function
        self.proj_fun = proj_fun

    def optimize(self):
        #  Initialize population
        if isinstance(self.start, np.ndarray):
            assert self.start.shape[0] == self.nindividuals and self.start.shape[1] == self.nvariables
            assert all(np.min(self.start, axis=0) >= self.lower_boundary) and \
                all(np.max(self.start, axis=0) <= self.upper_boundary)
            population = self.start
        elif self.start == "SLHD":
            exp_des = SymmetricLatinHypercube(self.nvariables, self.nindividuals)
            population = self.lower_boundary + exp_des.generate_points() * \
                (self.upper_boundary - self.lower_boundary)
        elif self.start == "LHD":
            exp_des = LatinHypercube(self.nvariables, self.nindividuals)
            population = self.lower_boundary + exp_des.generate_points() * \
                (self.upper_boundary - self.lower_boundary)
        elif self.start == "Random":
            population = self.lower_boundary + np.random.rand(self.nindividuals, self.nvariables) *\
                (self.upper_boundary - self.lower_boundary)
        else:
            raise AttributeError("Unknown argument for initial population")

        new_population = []
        #  Round positions
        if len(self.integer_variables) > 0:
            new_population = np.copy(population)
            population[:, self.integer_variables] = np.round(population[:, self.integer_variables])
            for i in self.integer_variables:
                ind = np.where(population[:, i] < self.lower_boundary[i])
                population[ind, i] += 1
                ind = np.where(population[:, i] > self.upper_boundary[i])
                population[ind, i] -= 1

        #  Evaluate all individuals
        function_values = self.function(population)
        if len(function_values.shape) == 2:
            function_values = np.squeeze(np.asarray(function_values))

        # Save the best individual
        ind = np.argmin(function_values)
        best_individual = np.copy(population[ind, :])
        best_value = function_values[ind]

        if len(self.integer_variables) > 0:
            population = new_population

        # Main loop
        for ngen in range(self.ngenerations):
            # Do tournament selection to select the parents
            competitors = np.random.randint(0, self.nindividuals, (self.nindividuals, self.tournament_size))
            ind = np.argmin(function_values[competitors], axis=1)
            winner_indices = np.zeros(self.nindividuals, dtype=int)
            for i in range(self.tournament_size):  # This loop is short
                winner_indices[np.where(ind == i)] = competitors[np.where(ind == i), i]

            parent1 = population[winner_indices[0:self.nindividuals/2], :]
            parent2 = population[winner_indices[self.nindividuals/2:self.nindividuals], :]

            # Averaging Crossover
            cross = np.where(np.random.rand(self.nindividuals/2) < self.p_cross)[0]
            nn = len(cross)  # Number of crossovers
            alpha = np.random.rand(nn, 1)

            # Create the new chromosomes
            parent1_new = np.multiply(alpha, parent1[cross, :]) + np.multiply(1-alpha, parent2[cross, :])
            parent2_new = np.multiply(alpha, parent2[cross, :]) + np.multiply(1-alpha, parent1[cross, :])
            parent1[cross, :] = parent1_new
            parent2[cross, :] = parent2_new
            population = np.concatenate((parent1, parent2))

            # Apply mutation
            scale_factors = self.sigma * (self.upper_boundary - self.lower_boundary)  # Account for dimensions ranges
            perturbation = np.random.randn(self.nindividuals, self.nvariables)  # Generate perturbations
            perturbation = np.multiply(perturbation, scale_factors)  # Scale accordingly
            perturbation = np.multiply(perturbation, (np.random.rand(self.nindividuals,
                                                                     self.nvariables) < self.p_mutation))

            population += perturbation  # Add perturbation
            population = np.maximum(np.reshape(self.lower_boundary, (1, self.nvariables)), population)
            population = np.minimum(np.reshape(self.upper_boundary, (1, self.nvariables)), population)

            # Map to feasible region if method exists
            if self.proj_fun is not None:
                for i in range(self.nindividuals):
                    population[i, :] = self.proj_fun(population[i, :])

            # Round chromosomes
            new_population = []
            if len(self.integer_variables) > 0:
                new_population = np.copy(population)
                population[:, self.integer_variables] = np.round(population[:, self.integer_variables])
                for i in self.integer_variables:
                    ind = np.where(population[:, i] < self.lower_boundary[i])
                    population[ind, i] += 1
                    ind = np.where(population[:, i] > self.upper_boundary[i])
                    population[ind, i] -= 1

            # Keep the best individual
            population[0, :] = best_individual

            #  Evaluate all individuals
            function_values = self.function(population)
            if len(function_values.shape) == 2:
                function_values = np.squeeze(np.asarray(function_values))

            # Save the best individual
            ind = np.argmin(function_values)
            best_individual = np.copy(population[ind, :])
            best_value = function_values[ind]

            # Use the positions that are not rounded
            if len(self.integer_variables) > 0:
                population = new_population

        return best_individual, best_value


def main():
    dim = 30

    # Vectorized Ackley function in dim dimensions
    def obj_function(x):
        return -20.0*np.exp(-0.2*np.sqrt(np.sum(x**2, axis=1)/dim)) - \
            np.exp(np.sum(np.cos(2.0*np.pi*x), axis=1)/dim)  + 20 + np.exp(1)

    ga = GeneticAlgorithm(obj_function, dim, -15*np.ones(dim), 20*np.ones(dim),
                          popsize=100, ngen=100, start="SLHD")
    x_best, f_best = ga.optimize()

    # Print the best solution found
    print("\nBest function value: {0}".format(f_best))
    print("Best solution: {0}".format(x_best))

    #  Add constraint of unit-1 norm and supply projection method

    def projection(x):
        return x / np.linalg.norm(x)

    ga = GeneticAlgorithm(obj_function, dim, -1*np.ones(dim), 1*np.ones(dim), popsize=100,
                          ngen=100, start="SLHD", proj_fun=projection)
    x_best, f_best = ga.optimize()

    # Print the best solution found
    print("\nBest function value: {0}".format(f_best))
    print("Best solution: {0}".format(x_best))
    print("norm(x_best) = {0}".format(np.linalg.norm(x_best)))

if __name__ == "__main__":
    main()




"""
.. module:: poly_regression
   :synopsis: Multivariate polynomial regression surface
.. moduleauthor:: David Bindel <bindel@cornell.edu>

:Module: poly_regression
:Author: David Bindel <bindel@cornell.edu>
"""

import numpy as np
import numpy.linalg as la


class PolyRegression(object):
    """Compute and evaluate a polynomial regression surface.

    :ivar nump: Current number of points
    :ivar maxp: Initial maximum number of points (can grow)
    :ivar x: Interpolation points
    :ivar fx: Function evaluations of interpolation points
    :ivar dim: Number of dimensions
    :ivar bounds: Upper and lower bounds, one row per dimension
    :ivar basisp: Multi-indices representing terms in a tensor poly basis
        Each row is a list of dim indices indicating a polynomial degree
        in the associated dimension.
    """

    def __init__(self, bounds, basisp, maxp=100):
        """Initialize the objects

        :param bounds: a (dims, 2) array of lower and upper bounds in each coord
        :param basisp: a (nbasis, dims) array, where the ith basis function is
            prod_j L_basisp(i,j)(x_j), L_k = the degree k Legendre polynomial
        """
        self.nump = 0
        self.maxp = maxp
        self.x = None     # pylint: disable=invalid-name
        self.fx = None
        self.bounds = bounds
        self.basisp = basisp
        self.updated = False

    def reset(self):
        """Reset the interpolation."""
        self.nump = 0
        self.x = None
        self.fx = None
        self.updated = False

    # FIXME
    def poised_for_interpolation(self, X):
        return True

    @property
    def dim(self):
        return self.bounds.shape[0]

    def _normalize(self, x):
        """Normalize points to the box [-1,1]^d
        """
        xx = np.copy(x)
        for k in range(x.shape[1]):
            l = self.bounds[k, 0]
            u = self.bounds[k, 1]
            w = u-l
            xx[:, k] = (x[:, k]-l)/w + (x[:, k]-u)/w
        return xx
        
    def _alloc(self):
        """Allocate storage for x and fx.
        """
        maxp = self.maxp
        self.x = np.zeros((maxp, self.dim))
        self.fx = np.zeros((maxp, 1))

    def _realloc(self, extra=1):
        """Expand allocation to accommodate more points (if needed)

        :param extra: Number of additional points to accommodate
        """
        if self.nump == 0:
            self._alloc()
        elif self.nump + extra > self.maxp:
            self.maxp = max(self.maxp*2, self.maxp + extra)
            self.x.resize((self.maxp, self.dim))
            self.fx.resize((self.maxp, 1))

    def _plegendre(self, x):
        """Evaluate basis functions.

        :param x: Coordinates (one per row)
        :return: Basis functions for each coordinate with shape (npts,nbasis)
        """
        s = self.basisp
        Px = legendre(x, np.max(s))
        Ps = np.ones((x.shape[0], s.shape[0]))
        for i in range(s.shape[0]):
            for j in range(s.shape[1]):
                Ps[:, i] *= Px[:, j, s[i, j]]
        return Ps

    def _dplegendre(self, x):
        """Evaluate basis function gradients.

        :param x: Coordinates (one per row)
        :return: Gradients for each coordinate with shape (npts,dim,nbasis)
        """
        s = self.basisp
        Px, dPx = dlegendre(x, np.max(s))
        dPs = np.ones((x.shape[0], x.shape[1], s.shape[0]))
        for i in range(s.shape[0]):
            for j in range(s.shape[1]):
                for k in range(x.shape[1]):
                    if k == j:
                        dPs[:, k, i] *= dPx[:, j, s[i, j]]
                    else:
                        dPs[:, k, i] *= Px[:, j, s[i, j]]
        return dPs

    def _fit(self):
        """Compute a least squares fit."""
        A = self._plegendre(self._normalize(self.get_x()))
        self.beta = la.lstsq(A, self.get_fx())[0]

    def _predict(self, xx):
        """Evaluate on response surface.
        """
        return np.dot(self._plegendre(self._normalize(xx)), self.beta)

    def _predict_deriv(self, xx):
        """Predict derivative.
        """
        dfx = np.dot(self._dplegendre(self._normalize(xx)), self.beta)
        for j in range(xx.shape[1]):
            dfx[:, j] /= (self.bounds[j, 1]-self.bounds[j, 0])/2
        return dfx

    def get_x(self):
        """Get the list of data points

        :return: List of data points
        """
        return self.x[:self.nump, :]

    def get_fx(self):
        """Get the list of function values for the data points.

        :return: List of function values
        """
        return self.fx[:self.nump, :]

    def add_point(self, xx, fx):
        """Add a new function evaluation

        :param xx: Point to add
        :param fx: The function value of the point to add
        """
        self._realloc()
        self.x[self.nump, :] = xx
        self.fx[self.nump, :] = fx
        self.nump += 1
        self.updated = False

    def eval(self, xx, d=None):
        """Evaluate the regression surface at point xx

        :param xx: Point where to evaluate
        :return: Value of the polynomial at xx
        """
        if self.updated is False:
            self._fit()
        self.updated = True

        xx = np.expand_dims(xx, axis=0)
        fx = self._predict(xx)
        return fx[0]

    def evals(self, xx, d=None):
        """Evaluate the regression surface at points xx

        :param xx: Points where to evaluate
        :return: Values of the polynomial at xx
        """
        if self.updated is False:
            self._fit()
        self.updated = True

        return np.atleast_2d(self._predict(xx))

    def deriv(self, x):
        """Evaluate the derivative of the regression surface at x

        :param x: Data point
        :return: Derivative of the polynomial at x
        """

        if self.updated is False:
            self._fit()
        self.updated = True

        x = np.expand_dims(x, axis=0)
        dfx = self._predict_deriv(x)
        return dfx[0]


def legendre(x, d):
    """Evaluate Legendre polynomials at all coordinates in x.

    :param x: Array of coordinates
    :param d: Max degree of polynomials
    :return: A x.shape-by-d array of Legendre polynomial values
    """
    x = np.array(x)
    s = x.shape + (d+1,)
    x = np.ravel(x)
    P = np.zeros((x.shape[0], d+1))
    P[:, 0] = 1
    if d > 0:
        P[:, 1] = x
    for n in range(1,d):
        P[:, n+1] = ((2*n+1)*(x*P[:, n]) - n*P[:, n-1])/(n+1)
    return P.reshape(s)


def dlegendre(x, d):
    """Evaluate Legendre polynomial derivatives at all coordinates in x.

    :param x: Array of coordinates
    :param d: Max degree of polynomials
    :return: x.shape-by-d arrays of Legendre polynomial values and derivatives
    """
    x = np.array(x)
    s = x.shape + (d+1,)
    x = np.ravel(x)
    P = np.zeros((x.shape[0], d+1))
    dP = np.zeros((x.shape[0], d+1))
    P[:, 0] = 1
    if d > 0:
        P[:, 1] = x
        dP[:, 1] = 1
    for n in range(1,d):
        P[:, n+1] = ((2*n+1)*(x*P[:, n]) - n*P[:, n-1])/(n+1)
        dP[:, n+1] = ((2*n+1)*(P[:, n] + x*dP[:, n]) - n*dP[:, n-1])/(n+1)
    return P.reshape(s), dP.reshape(s)


def basis_base(n, testf):
    """Generate list of shape functions for a subset of a TP poly space.

    :param n: Dimension of the space
    :param testf: Return True if a given multi-index is in range
    :return: An N-by-n matrix with S(i,j) = degree of variable j in shape i
    """
    snext = np.zeros((n,), dtype=np.int32)
    done = False

    # Follow carry chain through
    s = []
    while not done:
        s.append(snext.copy())
        done = True
        for i in range(n):
            snext[i] += 1
            if testf(snext):
                done = False
                break
            snext[i] = 0
    return np.array(s)


def basis_TP(n, d):
    """Generate list of shape functions for TP poly space.

    :param n: Dimension of the space
    :param d: Degree bound
    :return: An N-by-n matrix with S(i,j) = degree of variable j in shape i
           There are N = n^d shapes.
    """
    return basis_base(n, lambda s: np.all(s <= d))


def basis_TD(n, d):
    """Generate list of shape functions for TP poly space.

    :param n: Dimension of the space
    :param d: Degree bound
    :return: An N-by-n matrix with S(i,j) = degree of variable j in shape i
    """
    return basis_base(n, lambda s: np.sum(s) <= d)


def basis_HC(n, d):
    """Generate list of shape functions for HC poly space.

    :param n: Dimension of the space
    :param d: Degree bound
    :return: An N-by-n matrix with S(i,j) = degree of variable j in shape i
    """
    return basis_base(n, lambda s: np.prod(s+1) <= d+1)


def basis_SM(n, d):
    """Generate list of shape functions for SM poly space.

    :param n: Dimension of the space
    :param d: Degree bound
    :return: An N-by-n matrix with S(i,j) = degree of variable j in shape i
    """
    def fSM(p):
        return p if p < 2 else np.ceil(np.log2(p))

    def fSMv(s):
        f = 0
        for j in range(s.shape[0]):
            f += fSM(s[j])
        return f

    return basis_base(n, lambda s: fSMv(s) <= fSM(d))

# ====================================================================


def test_legendre1():
    npt = 1001
    x = np.linspace(-1,1, npt)
    w = np.ones((npt,1))/(npt-1)
    w[0] /= 2
    w[-1] /= 2
    P = legendre(x, 4)
    M = np.dot(P.T, w*P)
    E = M-np.diag(1/(2*np.arange(5)+1))
    relerr = la.norm(E)/la.norm(M)
    assert relerr < 1e-4, "Test Legendre orthonormality"


def test_legendre2():
    npt = 1000
    x = np.linspace(-1,1, npt)
    h = 2.0/(npt-1)
    P, dP = dlegendre(x, 4)
    assert np.max(np.abs(dP[1:-2,:]-(P[2:-1,:]-P[0:-3,:])/2/h)) < 1e-4, \
        "Test Legendre derivs vs finite difference"


def test_poly():
    bounds = np.array([[0, 1], [1, 1.5]])
    basisp = basis_TD(2, 2)
    surf = PolyRegression(bounds, basisp)

    def ref_poly(xy):
        x = xy[:,0]
        y = xy[:,1]
        return x**2 + y**2/3.0 + x*y/4.0 + x/5.0 + y/6.0 + 1.0

    def dref_poly(xy):
        x = xy[:,0]
        y = xy[:,1]
        return np.array([2.0*x + y/4.0 + 1/5.0,
                         2.0*y/3.0 + x/4.0 + 1/6.0])

    def add_point(x, y):
        xy = np.array([[x, y]])
        surf.add_point(xy, ref_poly(xy))

    add_point(0, 1)
    add_point(1, 1.5)
    add_point(0, 1.4)
    add_point(0.7, 1.1)
    add_point(0.5, 1.3)
    add_point(0.9, 1.2)
    add_point(0.1, 1.25)

    xtest = np.array([[0.123,1.456]])
    sxtest = surf.evals(xtest)
    fxtest = ref_poly(xtest)
    relerr = np.abs((sxtest-fxtest)/fxtest)
    assert relerr[0] < 1e-12

    dsxtest = surf.deriv(xtest[0,:])
    dfxtest = dref_poly(xtest)
    relerr = np.abs((dsxtest-dfxtest)/dfxtest)
    assert np.max(relerr) < 1e-12
    
if __name__ == "__main__":
    test_legendre1()
    test_legendre2()
    test_poly()




"""
.. module:: test_subprocess
  :synopsis: Test an external objective function
.. moduleauthor:: David Eriksson <dme65@cornell.edu>
"""

import logging
from pySOT import *
from poap.controller import ThreadController, ProcessWorkerThread
import numpy as np
from subprocess32 import Popen, PIPE
import os.path


def array2str(x):
    return ",".join(np.char.mod('%f', x))


class CppSim(ProcessWorkerThread):

    def handle_eval(self, record):
        self.process = Popen(['./sphere_ext', array2str(record.params[0])], stdout=PIPE)
        val = np.nan
        while True:
            output = self.process.stdout.readline()
            if output == '' and self.process.poll() is not None:  # No new output
                break
            if output:  # New intermediate output
                try:
                    val = float(output.strip())  # Try to parse output
                except ValueError:  # If the output is nonsense we ignore it
                    pass

        rc = self.process.poll()  # Check the return code
        if rc < 0 or np.isnan(val):
            logging.warning("Incorrect output or crashed evaluation")
            self.finish_cancelled(record)
        else:
            self.finish_success(record, val)


def main():
    if not os.path.exists("./logfiles"):
        os.makedirs("logfiles")
    if os.path.exists("./logfiles/test_subprocess.log"):
        os.remove("./logfiles/test_subprocess.log")
    logging.basicConfig(filename="./logfiles/test_subprocess.log",
                        level=logging.INFO)

    print("\nNumber of threads: 4")
    print("Maximum number of evaluations: 200")
    print("Search strategy: Candidate DyCORS")
    print("Experimental design: Latin Hypercube")
    print("Surrogate: Cubic RBF")

    assert os.path.isfile("./sphere_ext"), "You need to build sphere_ext"

    nthreads = 4
    maxeval = 200
    nsamples = nthreads

    data = Sphere(dim=10)
    print(data.info)

    # Create a strategy and a controller
    controller = ThreadController()
    controller.strategy = \
        SyncStrategyNoConstraints(
            worker_id=0, data=data,
            maxeval=maxeval, nsamples=nsamples,
            exp_design=LatinHypercube(dim=data.dim, npts=2*(data.dim+1)),
            sampling_method=CandidateDYCORS(data=data, numcand=100*data.dim),
            response_surface=RBFInterpolant(surftype=CubicRBFSurface, maxp=maxeval))

    # Launch the threads and give them access to the objective function
    for _ in range(nthreads):
        controller.launch_worker(CppSim(controller))

    # Run the optimization strategy
    result = controller.run()

    print('Best value found: {0}'.format(result.value))
    print('Best solution found: {0}\n'.format(
        np.array_str(result.params[0], max_line_width=np.inf,
                     precision=5, suppress_small=True)))

if __name__ == '__main__':
    main()



"""
.. module:: test_penalty
  :synopsis: Test constrained optimization strategy
.. moduleauthor:: David Eriksson <dme65@cornell.edu>
"""

import logging
from pySOT import *
from poap.controller import ThreadController, BasicWorkerThread
import numpy as np
import os.path


def main():
    if not os.path.exists("./logfiles"):
        os.makedirs("logfiles")
    if os.path.exists("./logfiles/test_penalty.log"):
        os.remove("./logfiles/test_penalty.log")
    logging.basicConfig(filename="./logfiles/test_penalty.log",
                        level=logging.INFO)

    print("\nNumber of threads: 4")
    print("Maximum number of evaluations: 500")
    print("Search strategy: CandidateDYCORS")
    print("Experimental design: Latin Hypercube")
    print("Surrogate: Cubic RBF")

    nthreads = 4
    maxeval = 500
    penalty = 1e6
    nsamples = nthreads

    data = Keane(dim=10)
    print(data.info)

    # Create a strategy and a controller
    controller = ThreadController()
    controller.strategy = \
        SyncStrategyPenalty(
            worker_id=0, data=data,
            maxeval=maxeval, nsamples=nsamples,
            response_surface=RBFInterpolant(surftype=CubicRBFSurface, maxp=maxeval),
            exp_design=LatinHypercube(dim=data.dim, npts=2*(data.dim+1)),
            sampling_method=CandidateDYCORS(data=data, numcand=100*data.dim),
            penalty=penalty)

    # Launch the threads
    for _ in range(nthreads):
        worker = BasicWorkerThread(controller, data.objfunction)
        controller.launch_worker(worker)

    # Use penalty based merit
    def feasible_merit(record):
        xx = np.zeros((1, record.params[0].shape[0]))
        xx[0, :] = record.params[0]
        return record.value + controller.strategy.penalty_fun(xx)[0, 0]
    result = controller.run(merit=feasible_merit)
    best, xbest = result.value, result.params[0]

    print('Best value: {0}'.format(best))
    print('Best solution: {0}'.format(
        np.array_str(xbest, max_line_width=np.inf,
                     precision=5, suppress_small=True)))
    print('Feasible: {0}\n'.format(np.max(data.eval_ineq_constraints(xbest)) <= 0.0))

if __name__ == '__main__':
    main()



"""
.. module:: ensemble_surrogate
   :synopsis: Ensemble surrogate surfaces
.. moduleauthor:: David Eriksson <dme65@cornell.edu>

:Module: ensemble_surrogate
:Author: David Eriksson <dme65@cornell.edu>
"""

from pyds import MassFunction
import numpy as np
from copy import copy, deepcopy
import math
import numpy.linalg as la


class EnsembleSurrogate:
    """Compute and evaluate an ensemble of interpolants.

    Maintains a list of surrogates and decides how to weights them
    by using Dempster-Shafer theory to assign pignistic probabilities
    based on statistics computed using LOOCV.

    :ivar nump: Current number of points
    :ivar maxp: Initial maximum number of points (can grow)
    :ivar rhs: Right hand side for interpolation system
    :ivar x: Interpolation points
    :ivar fx: Values at interpolation points
    :ivar dim: Number of dimensions
    :ivar model_list: List of surrogate models
    :ivar weights: Weight for each surrogate model
    :ivar surrogate_list: List of internal surrogate models for LOOCV
    """
    def __init__(self, model_list, maxp=100):
        self.nump = 0
        self.maxp = maxp
        self.x = None     # pylint: disable=invalid-name
        self.fx = None
        self.dim = None
        assert len(model_list) >= 2, "I need at least two models"
        self.model_list = model_list
        self.M = len(model_list)
        for i in range(self.M):
            self.model_list[i].reset()  # Models must be empty
        self.weights = None
        self.surrogate_list = None

    def reset(self):
        """Reset the interpolation."""
        self.nump = 0
        self.x = None
        self.fx = None
        for i in range(len(self.model_list)):
            self.model_list[i].reset()
        self.surrogate_list = None
        self.weights = None

    def _alloc(self, dim):
        """Allocate storage for x, fx, surrogate_list

        :param dim: Number of dimensions
        """
        maxp = self.maxp
        self.dim = dim
        self.x = np.zeros((maxp, dim))
        self.fx = np.zeros((maxp, 1))
        self.surrogate_list = [
            [None for _ in range(maxp)] for _ in range(self.M)]

    def _realloc(self, dim, extra=1):
        """Expand allocation to accommodate more points (if needed)

        :param dim: Number of dimensions
        :param extra: Number of additional points to accommodate
        """
        if self.nump == 0:
            self._alloc(dim)
        elif self.nump + extra > self.maxp - 1:
            oldmaxp = self.maxp
            self.maxp = max([self.maxp*2, self.maxp + extra])
            self.x.resize((self.maxp, dim))
            self.fx.resize((self.maxp, 1))
            # Expand the surrogate lists
            for i in range(self.M):
                for _ in range(self.maxp - oldmaxp):
                    self.surrogate_list[i].append(None)

    def _prob_to_mass(self, prob):
        """Internal method for building a mass function from probabilities

        :param prob: List of probabilities

        :return: Mass function
        """
        dictlist = []
        for i in range(len(prob)):
            dictlist.append([str(i+1), prob[i]])
        return MassFunction(dict(dictlist))

    def _mean_squared_error(self, x, y):
        return np.sum((x - y) ** 2)/len(x)

    def _mean_abs_err(self, x, y):
        return np.sum(np.abs(x - y))/len(x)

    def compute_weights(self):
        """Compute mode weights

        Given n observations we use n surrogates built with n-1 of the points
        in order to predict the value at the removed point. Based on these n
        predictions we calculate three different statistics:
            - Correlation coefficient with true function values
            - Root mean square deviation
            - Mean absolute error

        Based on these three statistics we compute the model weights by
        applying Dempster-Shafer theory to first compute the pignistic
        probabilities, which are taken as model weights.

        :return: Model weights
        """
        # Do the leave-one-out experiments
        loocv = np.zeros((self.M, self.nump))
        for i in range(self.M):
            for j in range(self.nump):
                loocv[i, j] = self.surrogate_list[i][j].eval(self.x[j, :])

        # Compute the model characteristics
        corr_coeff = np.ones(self.M)
        for i in range(self.M):
            corr_coeff[i] = np.corrcoef(np.vstack(
                (loocv[i, :], self.get_fx().flatten())))[0, 1]

        root_mean_sq_err = np.ones(self.M)
        for i in range(self.M):
            root_mean_sq_err[i] = 1.0 / math.sqrt(
                self._mean_squared_error(self.get_fx().flatten(), loocv[i, :]))

        mean_abs_err = np.ones(self.M)
        for i in range(self.M):
            mean_abs_err[i] = 1.0 / self._mean_abs_err(
                self.get_fx().flatten(), loocv[i, :])

        # Make sure no correlations are negative
        corr_coeff[np.where(corr_coeff < 0.0)] = 0.0
        if np.max(corr_coeff) == 0.0:
            corr_coeff += 1.0

        # Normalize the test statistics
        corr_coeff /= np.sum(corr_coeff)
        root_mean_sq_err /= np.sum(root_mean_sq_err)
        mean_abs_err /= np.sum(mean_abs_err)

        # Create mass functions based on the model characteristics
        m1 = self._prob_to_mass(corr_coeff)
        m2 = self._prob_to_mass(root_mean_sq_err)
        m3 = self._prob_to_mass(mean_abs_err)

        # Compute pignistic probabilities from Dempster-Shafer theory
        pignistic = m1.combine_conjunctive([m2, m3]).to_dict()
        self.weights = np.ones(self.M)
        for i in range(self.M):
            self.weights[i] = pignistic.get(str(i+1))

    def get_x(self):
        """Get the list of data points

        :return: List of data points
        """
        return self.x[:self.nump, :]

    def get_fx(self):
        """Get the list of function values for the data points.

        :return: List of function values
        """
        return self.fx[:self.nump, :]

    def add_point(self, xx, fx):
        """Add a new function evaluation

        This function also updates the list of LOOCV surrogate models by cleverly
        just adding one point to n of the models. The scheme in which new models
        are built is illustrated below:

        2           1           1,2

        2,3         1,3         1,2         1,2,3

        2,3,4       1,3,4       1,2,4       1,2,3       1,2,3,4

        2,3,4,5     1,3,4,5     1,2,4,5     1,2,3,5     1,2,3,4     1,2,3,4,5

        :param xx: Point to add
        :param fx: The function value of the point to add
        """
        dim = len(xx)
        self._realloc(dim)
        self.x[self.nump, :] = xx
        self.fx[self.nump, :] = fx
        self.nump += 1
        # Update the leave-one-out models
        if self.nump == 2:
            for i in range(self.M):
                #  Add the first three models
                x0 = copy(self.x[0, :])
                x1 = copy(self.x[1, :])
                self.surrogate_list[i][0] = deepcopy(self.model_list[i])
                self.surrogate_list[i][0].add_point(x1, self.fx[1])
                self.surrogate_list[i][1] = deepcopy(self.model_list[i])
                self.surrogate_list[i][1].add_point(x0, self.fx[0])
                self.surrogate_list[i][2] = deepcopy(self.surrogate_list[i][1])
                self.surrogate_list[i][2].add_point(x1, self.fx[1])
        elif self.nump > 2:
            for i in range(self.M):
                for j in range(self.nump-1):
                    self.surrogate_list[i][j].add_point(xx, fx)
                self.surrogate_list[i][self.nump] = deepcopy(
                    self.surrogate_list[i][self.nump-1])
                self.surrogate_list[i][self.nump].add_point(xx, fx)
                # Point to the model with all points
                self.model_list[i] = self.surrogate_list[i][self.nump]
        self.weights = None

    def eval(self, xx, d=None):
        """Evaluate the interpolant at the point xx

        :param xx: Point where to evaluate

        :return: Value of the MARS interpolant at x
        """
        if self.weights is None:
            self.compute_weights()

        val = 0
        for i in range(self.M):
            val += self.weights[i]*self.model_list[i].eval(xx, d)
        return val

    def evals(self, xx, d=None):
        """Evaluate the MARS interpolant at the points xx

        :param xx: Points where to evaluate

        :return: Values of the MARS interpolant at x
        """
        if self.weights is None:
            self.compute_weights()

        vals = np.zeros((xx.shape[0], 1))
        for i in range(self.M):
            vals += self.weights[i] * self.model_list[i].evals(xx, d)

        return vals

    def deriv(self, x, d=None):
        """Evaluate the derivative of the MARS interpolant at x

        :param x: Data point

        :return: Derivative of the MARS interpolant at x
        """
        if self.weights is None:
            self.compute_weights()

        val = 0.0
        for i in range(self.M):
            val += self.weights[i]*self.model_list[i].deriv(x, d)
        return val

if __name__ == "__main__":

    from pySOT import RBFInterpolant
    from pySOT import CubicRBFSurface, TPSSurface, LinearRBFSurface

    fhat1 = RBFInterpolant(CubicRBFSurface, 1e-8, 100)
    fhat2 = RBFInterpolant(TPSSurface, 1e-8, 100)
    fhat3 = RBFInterpolant(LinearRBFSurface, 1e-8, 100)

    models = [fhat1, fhat2, fhat3]
    fhat = EnsembleSurrogate(models, 10)

    def test_f(x):
        """Test function"""
        fx = x[1]*np.sin(x[0]) + x[0]*np.cos(x[1])
        return fx

    def test_df(x):
        """Derivative of test function"""
        dfx = np.array([x[1]*np.cos(x[0])+np.cos(x[1]),
                        np.sin(x[0])-x[0]*np.sin(x[1])])
        return dfx

    xs = np.random.rand(120, 2)
    for ii in range(100):
        xx = xs[ii, :]
        fx = test_f(xx)
        fhat.add_point(xx, fx)
    fhx = fhat.evals(xs[:5, :])
    print("Weights: " + str(fhat.weights))
    for ii in range(5):
        fx = test_f(xs[ii, :])
        print("Err: %e" % (abs(fx-fhx[ii])/abs(fx)))
    for ii in range(10):
        xx = xs[100 + ii, :]
        fx = test_f(xx)
        dfx = test_df(xx)
        fhx = fhat.eval(xx)
        dfhx = fhat.deriv(xx)
        print("Err (interp): %e : %e" % (abs(fx-fhx)/abs(fx),
                                         la.norm(dfx-dfhx)/la.norm(dfx)))



"""
.. module:: search_procedure
  :synopsis: ways of finding the next point to evaluate
.. moduleauthor:: David Eriksson <dme65@cornell.edu>
                 David Bindel <bindel@cornell.edu>

:Module: sampling_strategies
:Author: David Eriksson <dme65@cornell.edu>,
    David Bindel <bindel@cornell.edu>

We currently only support the weighted distance merit function.

We support the following methods for minimizing on the surface:

Candidate based methods:
    - CandidateSRBF: Generate candidate points around the best point
    - CandidateDYCORS: Uses a DDS strategy with a cap on the lowest probability
    - CandidateUniform: Sample the candidate points uniformly in the domain
    - CandidateSRBF_INT: Uses CandidateSRBF but only
        perturbs the integer variables
    - CandidateSRBF_CONT: Uses CandidateSRBF but only
        perturbs the continuous variables
    - CandidateDYCORS_INT: Uses CandidateSRBF but only
        perturbs the integer variables
    - CandidateDYCORS_CONT: Uses CandidateSRBF but only
        perturbs the continuous variables
    - CandidateUniform_CONT: Sample the continuous variables of
        the candidate points uniformly in the domain
    - CandidateUniform_INT: Sample the integer variables of the
        candidate points uniformly in the domain

We also support using multiple of these strategies and cycle
    between them which we call MultiSearchStrategy
    - MultiSearchStrategy
"""

import math
from utils import *
import scipy.spatial as scp
from heuristic_algorithms import GeneticAlgorithm as GA
from scipy.optimize import minimize
import scipy.stats as stats
from merit_functions import *

# ========================= MultiSearch =======================


class MultiSampling(object):
    """ A collection of Sampling Methods and weights so that the user
        can use multiple sampling methods for the same optimization
        problem. This object keeps an internal list of proposed points
        in order to be able to compute the minimum distance from a point
        to all proposed evaluations. This list has to be reset each time
        the optimization algorithm restarts"""
    def __init__(self, strategy_list, cycle):
        if cycle is None:
            cycle = range(len(strategy_list))
        if (not all(isinstance(i, int) for i in cycle)) or \
                np.min(cycle) < 0 or \
                np.max(cycle) > len(strategy_list)-1:
            raise ValueError("Incorrect cycle!!")
        self.sampling_strategies = strategy_list
        self.nstrats = len(strategy_list)
        self.cycle = cycle
        self.current_strat= 0
        self.proposed_points = None
        self.data = strategy_list[0].data
        self.fhat = None
        self.avoid = None
        self.budget = None
        self.n0 = None

    def init(self, start_sample, fhat, budget):
        """Initilize the sampling method by providing the points in the
        experimental design, the surrogate model, and the evaluation budget

        :param start_sample: Points in the experimental design
        :param fhat: Surrogate model
        :param budget: Evaluation budget
        """
        self.proposed_points = start_sample
        self.fhat = fhat
        self.n0 = start_sample.shape[0]
        for i in range(self.nstrats):
            self.sampling_strategies[i].init(self.proposed_points, fhat, budget)

    def remove_point(self, x):
        """Remove x from proposed_points. Useful if x was never evaluated.

        :param x: Point to be removed

        :return: True if points was removed, False otherwise
        """
        idx = np.sum(np.abs(self.proposed_points - x), axis=1).argmin()
        if np.sum(np.abs(self.proposed_points[idx, :] - x)) < 1e-10:
            self.proposed_points = np.delete(self.proposed_points, idx, axis=0)
            for i in range(self.nstrats):
                self.sampling_strategies[i].remove_point(x)
            return True
        return False

    def make_points(self, npts, xbest, sigma, subset=None, proj_fun=None,
                    merit=candidate_merit_weighted_distance):
        """Create new candidate points. This call is ignored by the optimization
        based search strategies.

        :param npts: Number of points to select
        :param xbest: Best solution found so far
        :param sigma: Current sampling radius w.r.t the unit box
        :param subset: Coordinates to perturb
        :param proj_fun: Routine for projecting infeasible points onto the feasible region
        :param merit: Merit function for selecting candidate points

        :return: Points selected for evaluation"""

        new_points = np.zeros((npts, self.data.dim))

        # Figure out what we need to generate
        npoints = np.zeros((self.nstrats,), dtype=int)
        for i in range(npts):
            npoints[self.cycle[self.current_strat]] += 1
            self.current_strat = (self.current_strat + 1) % len(self.cycle)

        # Now generate the points from one strategy at the time
        count = 0
        for i in range(self.nstrats):
            if npoints[i] > 0:
                new_points[count:count+npoints[i], :] = \
                    self.sampling_strategies[i].make_points(npts=npoints[i], xbest=xbest,
                                                            sigma=sigma, subset=subset,
                                                            proj_fun=proj_fun,
                                                            merit=merit)

                count += npoints[i]
                # Update list of proposed points
                for j in range(self.nstrats):
                    if j != i:
                        self.sampling_strategies[j].proposed_points = \
                            self.sampling_strategies[i].proposed_points

        return new_points


class CandidateSRBF(object):
    """This is an implementation of the candidate points method that is
    proposed in the first SRBF paper. Candidate points are generated
    by making normally distributed perturbations with stdDev sigma
    around the best solution

    :ivar data: Optimization object
    :ivar weights: Weights used in the merit function
    :ivar numcand: Number of candidate points to generate
    :ivar proposed_points: List of points proposed by any search strategy
        since the last restart
    """

    def __init__(self, data, numcand=None, weights=None):
        """Initialize the SRBF method

        :param data: Optimization object
        :param numcand: Number of candidate points to generate
        :param weights: Weights used for the merit function
        """
        self.data = data
        self.fhat = None
        self.xrange = self.data.xup - self.data.xlow
        self.dtol = 1e-3 * math.sqrt(data.dim)
        self.weights = weights
        if self.weights is None:
            self.weights = [0.3, 0.5, 0.8, 0.95]
        self.proposed_points = None
        self.dmerit = None
        self.xcand = None
        self.fhvals = None
        self.next_weight = 0
        self.numcand = numcand
        if self.numcand is None:
            self.numcand = min([5000, 100*data.dim])
        self.budget = None
        self.n0 = None
        self.fhat = None

        # Check that the inputs make sense
        assert isinstance(self.numcand, int) and self.numcand > 0, \
            "The number of candidate points has to be a positive integer"
        assert (isinstance(self.weights, np.ndarray) or
                isinstance(self.weights, list)) and max(self.weights) <= 1 \
            and min(self.weights) >= 0, "Incorrect weights"

    def init(self, start_sample, fhat, budget):
        """Initialize the sampling method by providing the points in the
        experimental design, the surrogate model, and the evaluation budget

        :param start_sample: Points in the experimental design
        :param fhat: Surrogate model
        :param budget: Evaluation budget
        """
        self.proposed_points = start_sample
        self.n0 = start_sample.shape[0]
        self.budget = budget
        self.fhat = fhat

    def remove_point(self, x):
        """Remove x from the list of proposed points.
        Useful if x was never evaluated.

        :param x: Point to be removed

        :return: True if points was removed, False otherwise
        """
        idx = np.sum(np.abs(self.proposed_points - x), axis=1).argmin()
        if np.sum(np.abs(self.proposed_points[idx, :] - x)) < 1e-10:
            self.proposed_points = np.delete(self.proposed_points, idx, axis=0)
            return True
        return False

    def __generate_cand__(self, scalefactors, xbest, subset):
        self.xcand = np.ones((self.numcand,  self.data.dim)) * xbest
        for i in subset:
            lower, upper = self.data.xlow[i], self.data.xup[i]
            ssigma = scalefactors[i]
            self.xcand[:, i] = stats.truncnorm.rvs(
                (lower - xbest[i]) / ssigma, (upper - xbest[i]) / ssigma,
                loc=xbest[i], scale=ssigma, size=self.numcand)

    def make_points(self, npts, xbest, sigma, subset=None, proj_fun=None,
                    merit=candidate_merit_weighted_distance):
        """Create new candidate points based on the best solution and the current value of sigma.

        :param npts: Number of points to select
        :param xbest: Best solution found so far
        :param sigma: Current sampling radius w.r.t the unit box
        :param subset: Coordinates to perturb
        :param proj_fun: Routine for projecting infeasible points onto the feasible region
        :param merit: merit function for selecting candidate points

        :return: Points selected for evaluation
        """

        if subset is None:
            subset = np.arange(0, self.data.dim)
        scalefactors = sigma * self.xrange

        # Make sure that the scale factors are correct for
        # the integer variables (at least 1)
        ind = np.intersect1d(self.data.integer, subset)
        if len(ind) > 0:
            scalefactors[ind] = np.maximum(scalefactors[ind], 1.0)

        # Generate candidate points
        self.__generate_cand__(scalefactors, xbest, subset)
        if proj_fun is not None:
            self.xcand = proj_fun(self.xcand)

        dists = scp.distance.cdist(self.xcand, self.proposed_points)
        fhvals = self.fhat.evals(self.xcand)

        self.dmerit = np.amin(np.asmatrix(dists), axis=1)
        self.fhvals = unit_rescale(fhvals)

        xnew = merit(self, npts)
        self.proposed_points = np.vstack((self.proposed_points,
                                          np.asmatrix(xnew)))
        return xnew


class CandidateUniform(CandidateSRBF):
    """Create Candidate points by sampling uniformly in the domain"""

    def __generate_cand__(self, scalefactors, xbest, subset):
        self.xcand = np.ones((self.numcand, self.data.dim)) * xbest
        self.xcand[:, subset] = np.random.uniform(
            self.data.xlow[subset], self.data.xup[subset], (self.numcand, len(subset)))


class CandidateDYCORS(CandidateSRBF):
    """This is an implementation of DyCORS method to generate
    candidate points. The DyCORS method only perturbs a subset
    of the dimensions when perturbing the best solution. The
    probability for a dimension to be perturbed decreases after
    each evaluation and is capped in order to guarantee
    global convergence."""

    def __init__(self, data, numcand=None, weights=None):
        """Initialize the DYCORS method

        :param data: Optimization object
        :param numcand:  Number of candidate points to generate"""

        CandidateSRBF.__init__(self, data, numcand=numcand, weights=weights)
        self.minprob = np.min([1.0, 1.0/self.data.dim])
        assert data.dim > 1, "You can't use DYCORS on a 1d problem"

        def probfun(numevals, budget):
            if budget < 2:
                return 0
            return min([20.0/data.dim, 1.0]) * (1.0 - (np.log(numevals + 1.0) / np.log(budget)))
        self.probfun = probfun

    def __generate_cand__(self, scalefactors, xbest, subset):
        ddsprob = self.probfun(self.proposed_points.shape[0] - self.n0, self.budget - self.n0)
        ddsprob = np.max([self.minprob, ddsprob])

        nlen = len(subset)
        ar = (np.random.rand(self.numcand, nlen) < ddsprob)
        ind = np.where(np.sum(ar, axis=1) == 0)[0]
        ar[ind, np.random.randint(0, nlen - 1, size=len(ind))] = 1

        self.xcand = np.ones((self.numcand, self.data.dim)) * xbest
        for i in range(nlen):
            lower, upper = self.data.xlow[i], self.data.xup[i]
            ssigma = scalefactors[subset[i]]
            ind = np.where(ar[:, i] == 1)[0]
            self.xcand[ind, subset[i]] = stats.truncnorm.rvs(
                (lower - xbest[subset[i]]) / ssigma, (upper - xbest[subset[i]]) / ssigma,
                loc=xbest[subset[i]], scale=ssigma, size=len(ind))


class CandidateDDS(CandidateDYCORS):
    """This is an implementation of DDS method to generate
    candidate points. Only a few candidate points are generated
    and the candidate point with the lowest value predicted
    by the surrogate model is selected. The DDS method only
    perturbs a subset of the dimensions when perturbing the
    best solution. The probability for a dimension to be
    perturbed decreases after each evaluation and is capped
    in order to guarantee global convergence."""

    def __init__(self, data, numcand=None, weights=None):
        """Initialize the DDS method

        :param data: Optimization object
        :param numcand:  Number of candidate points to generate"""

        CandidateDYCORS.__init__(self, data, numcand=numcand, weights=weights)
        self.weights = np.array([1.0])
        self.numcand = max([0.5*data.dim, 2])

        def probfun(numevals, budget):
            return 1.0-(np.log(numevals + 1.0) / np.log(budget))
        self.probfun = probfun

    def make_points(self, npts, xbest, sigma, subset=None, proj_fun=None,
                    merit=candidate_merit_weighted_distance):
        """Create new candidate points based on the best solution and the current value of sigma.

        :param npts: Number of points to select
        :param xbest: Best solution found so far
        :param sigma: Current sampling radius w.r.t the unit box
        :param subset: Coordinates to perturb
        :param proj_fun: Routine for projecting infeasible points onto the feasible region
        :param merit: merit function for selecting candidate points

        :return: Points selected for evaluation"""

        new_points = np.zeros((npts, self.data.dim))
        for i in range(npts):
            new_points[i, :] = CandidateDYCORS.make_points(self, npts=1, xbest=xbest, sigma=0.2,
                                                           subset=subset, proj_fun=proj_fun)
        return new_points


class CandidateSRBF_INT(CandidateSRBF):
    """Candidate points are generated by perturbing ONLY the discrete
    variables using the SRBF strategy
    """
    def make_points(self, npts, xbest, sigma, subset=None, proj_fun=None,
                    merit=candidate_merit_weighted_distance):
        if len(self.data.integer) > 0:
            return CandidateSRBF.make_points(self, npts=npts, xbest=xbest, sigma=sigma,
                                             subset=self.data.integer, proj_fun=proj_fun,
                                             merit=merit)
        else:
            return CandidateSRBF.make_points(self, npts=npts, xbest=xbest, sigma=sigma,
                                             subset=self.data.continuous, proj_fun=proj_fun,
                                             merit=merit)


class CandidateDYCORS_INT(CandidateDYCORS):
    """Candidate points are generated by perturbing ONLY the discrete
    variables using the DYCORS strategy"""

    def make_points(self, npts, xbest, sigma, subset=None, proj_fun=None,
                    merit=candidate_merit_weighted_distance):
        if len(self.data.integer) > 0:
            return CandidateDYCORS.make_points(self, npts=npts, xbest=xbest, sigma=sigma,
                                               subset=self.data.integer, proj_fun=proj_fun,
                                               merit=merit)
        else:
            return CandidateDYCORS.make_points(self, npts=npts, xbest=xbest, sigma=sigma,
                                               subset=self.data.continuous, proj_fun=proj_fun,
                                               merit=merit)


class CandidateDDS_INT(CandidateDDS):
    """Candidate points are generated by perturbing
    ONLY the discrete variables using the DDS strategy"""

    def make_points(self, npts, xbest, sigma, subset=None, proj_fun=None,
                    merit=candidate_merit_weighted_distance):
        if len(self.data.integer) > 0:
            return CandidateDDS.make_points(self, npts=npts, xbest=xbest, sigma=sigma,
                                            subset=self.data.integer, proj_fun=proj_fun,
                                            merit=merit)
        else:
            return CandidateDDS.make_points(self, npts=npts, xbest=xbest, sigma=sigma,
                                            subset=self.data.continuous, proj_fun=proj_fun,
                                            merit=merit)


class CandidateUniform_INT(CandidateUniform):
    """Candidate points are generated by perturbing ONLY
    the discrete variables using the uniform perturbations"""

    def make_points(self, npts, xbest, sigma, subset=None, proj_fun=None,
                    merit=candidate_merit_weighted_distance):
        if len(self.data.integer) > 0:
            return CandidateUniform.make_points(self, npts=npts, xbest=xbest, sigma=sigma,
                                                subset=self.data.integer, proj_fun=proj_fun,
                                                merit=merit)
        else:
            return CandidateUniform.make_points(self, npts=npts, xbest=xbest, sigma=sigma,
                                                subset=self.data.continuous, proj_fun=proj_fun,
                                                merit=merit)


class CandidateSRBF_CONT(CandidateSRBF):
    """Candidate points are generated by perturbing ONLY
    the continuous variables using the SRBF strategy"""

    def make_points(self, npts, xbest, sigma, subset=None, proj_fun=None,
                    merit=candidate_merit_weighted_distance):
        if len(self.data.continuous) > 0:
            return CandidateSRBF.make_points(self, npts=npts, xbest=xbest, sigma=sigma,
                                             subset=self.data.continuous, proj_fun=proj_fun,
                                             merit=merit)
        else:
            return CandidateSRBF.make_points(self, npts=npts, xbest=xbest, sigma=sigma,
                                             subset=self.data.integer, proj_fun=proj_fun,
                                             merit=merit)


class CandidateDYCORS_CONT(CandidateDYCORS):
    """Candidate points are generated by perturbing ONLY
    the continuous variables using the DYCORS strategy"""

    def make_points(self, npts, xbest, sigma, subset=None, proj_fun=None,
                    merit=candidate_merit_weighted_distance):
        if len(self.data.continuous) > 0:
            return CandidateDYCORS.make_points(self, npts=npts, xbest=xbest, sigma=sigma,
                                               subset=self.data.continuous, proj_fun=proj_fun,
                                               merit=merit)
        else:
            return CandidateDYCORS.make_points(self, npts=npts, xbest=xbest, sigma=sigma,
                                               subset=self.data.integer, proj_fun=proj_fun,
                                               merit=merit)


class CandidateDDS_CONT(CandidateDDS):
    """Candidate points are generated by perturbing
    ONLY the discrete variables using the DDS strategy"""

    def make_points(self, npts, xbest, sigma, subset=None, proj_fun=None,
                    merit=candidate_merit_weighted_distance):
        if len(self.data.continuous) > 0:
            return CandidateDDS.make_points(self, npts=npts, xbest=xbest, sigma=sigma,
                                            subset=self.data.continuous, proj_fun=proj_fun,
                                            merit=merit)
        else:
            return CandidateDDS.make_points(self, npts=npts, xbest=xbest, sigma=sigma,
                                            subset=self.data.integer, proj_fun=proj_fun,
                                            merit=merit)


class CandidateUniform_CONT(CandidateUniform):
    """Candidate points are generated by perturbing ONLY
    the continuous variables using uniform perturbations"""

    def make_points(self, npts, xbest, sigma, subset=None, proj_fun=None,
                    merit=candidate_merit_weighted_distance):
        if len(self.data.continuous) > 0:
            return CandidateUniform.make_points(self, npts=npts, xbest=xbest, sigma=sigma,
                                                subset=self.data.continuous, proj_fun=proj_fun,
                                                merit=merit)
        else:
            return CandidateUniform.make_points(self, npts=npts, xbest=xbest, sigma=sigma,
                                                subset=self.data.integer, proj_fun=proj_fun,
                                                merit=merit)


################################## Optimization based strategies ###################################


class GeneticAlgorithm(object):
    def __init__(self, data):
        """Initialize the Genetic Algorithm method

        :param data: Optimization object
        """

        self.data = data
        self.fhat = None
        self.dtol = 1e-3 * math.sqrt(data.dim)
        self.proposed_points = None
        self.budget = None
        self.fhat = None

    def init(self, start_sample, fhat, budget):
        """Initialize the sampling method by providing the points in the
        experimental design, the surrogate model, and the evaluation budget

        :param start_sample: Points in the experimental design
        :param fhat: Surrogate model
        :param budget: Evaluation budget
        """

        self.proposed_points = start_sample
        self.budget = budget
        self.fhat = fhat

    def remove_point(self, x):
        """Remove x from the list of proposed points.
        Useful if x was never evaluated.

        :param x: Point to be removed

        :return: True if points was removed, False otherwise
        """

        idx = np.sum(np.abs(self.proposed_points - x), axis=1).argmin()
        if np.sum(np.abs(self.proposed_points[idx, :] - x)) < 1e-10:
            self.proposed_points = np.delete(self.proposed_points, idx, axis=0)
            return True
        return False

    def make_points(self, npts, xbest, sigma, subset=None, proj_fun=None, merit=None):
        """Create new points to evaluate using the GA

        :param npts: Number of points to select
        :param xbest: Best solution found so far (ignored)
        :param sigma: Current sampling radius w.r.t the unit box (ignored)
        :param subset: Coordinates to perturb
        :param proj_fun: Routine for projecting infeasible points onto the feasible region
        :param merit: Merit function for selecting candidate points (ignored)

        :return: Points selected for evaluation"""

        new_points = np.zeros((npts, self.data.dim))
        for i in range(npts):
            ga = GA(self.fhat.evals, self.data.dim, self.data.xlow, self.data.xup,
                    popsize=max([2*self.data.dim, 100]), ngen=100, proj_fun=proj_fun)
            x_min, f_min = ga.optimize()

            dist = np.atleast_2d(np.min(
                scp.distance.cdist(self.proposed_points, np.atleast_2d(x_min)), axis=1)).T

            x_new = x_min
            if np.min(dist) < self.dtol:
                # Perturb the best solution until we satisfy the tolerance
                d = 0.0
                x_new = None
                while d < self.dtol:
                    x_new = x_min + self.dtol * np.random.randn(1, self.data.dim)
                    x_new = np.maximum(x_new, self.data.xlow)
                    x_new = np.minimum(x_new, self.data.xup)
                    d = np.atleast_2d(np.min(
                        scp.distance.cdist(self.proposed_points, np.atleast_2d(x_new)), axis=1)).T.min()

            new_points[i, :] = x_new
            self.proposed_points = np.vstack((self.proposed_points, np.asarray(x_new)))

        return new_points


class MultiStartGradient(object):
    """ A wrapper around the scipy.optimize implementations of box-constrained
        gradient based minimization.

    Attributes:
        usecand: Indicates that this method is NOT candidate based
        data: Optimization object
        fhat: Original response surface object.
        proposed_points: List of points proposed by any search strategy
                         since the last restart
        objfun: The merit function to minimize (needs to have a gradient available)
        bounds: n x 2 matrix with lower and upper bound constraints
        numrestarts: Number of random starting points
        method: What optimization method to use. The following options are available:
            - L-BFGS-B: Quasi-Newton method of
                        Broyden, Fletcher, Goldfarb, and Shanno (BFGS)
            - TNC:      Truncated Newton algorithm

        Note: SLSQP is supposed to work with bound constraints but for some reason it
              sometimes violates the constraints anyway.
    """

    def __init__(self, data, method='L-BFGS-B', num_restarts=30):
        """Initialize the Multi-Start Gradient object

        :param data: Optimization object
        :param method: Optimization method to use for the surrogate
        :param num_restarts: Number of restarts
        """

        self.data = data
        self.fhat = None
        self.bounds = np.zeros((self.data.dim, 2))
        self.bounds[:, 0] = self.data.xlow
        self.bounds[:, 1] = self.data.xup
        self.eval = None
        self.deriv = None
        self.dtol = 1e-3 * math.sqrt(data.dim)
        self.proposed_points = None
        self.budget = None
        self.num_restarts = num_restarts
        self.x_best = None
        if (method == 'TNC') or (method == 'L-BFGS-B'):
            self.method = method
        else:
            self.method = 'L-BFGS-B'

    def init(self, start_sample, fhat, budget):
        """Initialize the sampling method by providing the points in the
        experimental design, the surrogate model, and the evaluation budget

        :param start_sample: Points in the experimental design
        :param fhat: Surrogate model
        :param budget: Evaluation budget
        """

        self.proposed_points = start_sample
        self.budget = budget
        self.fhat = fhat

    def remove_point(self, x):
        """Remove x from the list of proposed points.
        Useful if x was never evaluated.

        :param x: Point to be removed

        :return: True if points was removed, False otherwise
        """

        idx = np.sum(np.abs(self.proposed_points - x), axis=1).argmin()
        if np.sum(np.abs(self.proposed_points[idx, :] - x)) < 1e-10:
            self.proposed_points = np.delete(self.proposed_points, idx, axis=0)
            return True
        return False

    def make_points(self, npts, xbest, sigma, subset=None, proj_fun=None, merit=None):
        """Create new points to evaluate using the Multi-Start Gradient method

        :param npts: Number of points to select
        :param xbest: Best solution found so far (ignored)
        :param sigma: Current sampling radius w.r.t the unit box (ignored)
        :param subset: Coordinates to perturb
        :param proj_fun: Routine for projecting infeasible points onto the feasible region
        :param merit: Merit for selecting candidate points (ignored)

        :return: Points selected for evaluation"""

        def eval(x):
            return self.fhat.eval(x).ravel()

        def deriv(x):
            return self.fhat.deriv(x).ravel()

        new_points = np.zeros((npts, self.data.dim))
        for j in range(npts):
            fvals = np.zeros(self.num_restarts)
            xvals = np.zeros((self.num_restarts, self.data.dim))
            dists = np.zeros(self.num_restarts)

            for i in range(self.num_restarts):
                if i == 0 and j == 0:
                    x0 = np.array(xbest)
                else:
                    x0 = np.random.uniform(self.data.xlow, self.data.xup)

                res = minimize(eval, x0, method=self.method,
                               jac=deriv, bounds=self.bounds)

                # Compute the distance to the proposed points
                xx = np.atleast_2d(res.x)
                if proj_fun is not None:
                    xx = proj_fun(xx)
                dist = np.atleast_2d(np.min(
                    scp.distance.cdist(self.proposed_points, xx), axis=1)).T

                fvals[i] = res.fun
                xvals[i, :] = xx
                dists[i] = dist.min()

            if dists.max() > self.dtol:
                x_new = None
                f_new = np.inf
                for i in range(self.num_restarts):
                    if dists[i] > self.dtol and fvals[i] < f_new:
                        x_new = xvals[i, :]
                        f_new = fvals[i]
            else:
                # Perturb the best point
                d = -1.0
                x_new = None
                while d < self.dtol:
                    x_new = xvals[np.argmin(fvals), :] + self.dtol * np.random.randn(1, self.data.dim)
                    x_new = np.maximum(x_new, self.data.xlow)
                    x_new = np.minimum(x_new, self.data.xup)
                    if proj_fun is not None:
                        x_new = proj_fun(x_new)
                    d = np.atleast_2d(np.min(
                        scp.distance.cdist(self.proposed_points, np.atleast_2d(x_new)), axis=1)).T.min()

            new_points[j, :] = x_new
            self.proposed_points = np.vstack((self.proposed_points, np.asarray(x_new)))

        return new_points



"""
.. module:: experimental_design
  :synopsis: Methods for generating an experimental design.
.. moduleauthor:: David Eriksson <dme65@cornell.edu>,
                 Yi Shen <ys623@cornell.edu>

:Module: experimental_design
:Author: David Eriksson <dme65@cornell.edu>,
    Yi Shen <ys623@cornell.edu>
"""

import numpy as np
import pyDOE as pydoe


class LatinHypercube(object):
    """Latin Hypercube experimental design

    :ivar dim: Number of dimensions
    :ivar npts: Number of desired sampling points
    :ivar criterion: A string that tells lhs how to sample the
        points (default: None which simply randomizes the points
        within the intervals):
            - "center" or "c": center the points within the sampling intervals
            - "maximin" or "m": maximize the minimum distance
                between points, but place the point in a randomized
                location within its interval
            - "centermaximin" or "cm": same as "maximin", but
                centered within the intervals
            - "correlation" or "corr": minimize the maximum
                correlation coefficient
    """

    def __init__(self, dim, npts, criterion='c'):
        self.dim = dim
        self.npts = npts
        self.criterion = criterion

    def generate_points(self):
        """Generate a matrix with the initial sample points,
        scaled to the unit cube

        :return: Latin hypercube design in the unit cube
        """
        return pydoe.lhs(self.dim, self.npts, self.criterion)


class SymmetricLatinHypercube(object):
    """Symmetric Latin Hypercube experimental design

    :ivar dim: Number of dimensions
    :ivar npts: Number of desired sampling points
    """
    def __init__(self, dim, npts):
        self.dim = dim
        self.npts = npts

    def _slhd(self):
        """Generate matrix of sample points in the unit box"""

        # Generate a one-dimensional array based on sample number
        points = np.zeros([self.npts, self.dim])
        points[:, 0] = np.arange(1, self.npts+1)

        # Get the last index of the row in the top half of the hypercube
        middleind = self.npts//2

        # special manipulation if odd number of rows
        if self.npts % 2 == 1:
            points[middleind, :] = middleind + 1

        # Generate the top half of the hypercube matrix
        for j in range(1, self.dim):
            for i in range(middleind):
                if np.random.random() < 0.5:
                    points[i, j] = self.npts-i
                else:
                    points[i, j] = i + 1
            np.random.shuffle(points[:middleind, j])

        # Generate the bottom half of the hypercube matrix
        for i in range(middleind, self.npts):
            points[i, :] = self.npts + 1 - points[self.npts - 1 - i, :]

        return points/self.npts

    def generate_points(self):
        """Generate a matrix no rank deficiency with the initial
        sample points, scaled to the unit cube

        :return: Symmetric Latin hypercube design in the unit cube
        """
        rank_pmat = 0
        pmat = np.ones((self.npts, self.dim+1))
        xsample = None
        while rank_pmat != self.dim + 1:
            xsample = self._slhd()
            pmat[:, 1:] = xsample
            rank_pmat = np.linalg.matrix_rank(pmat)
        return xsample


class TwoFactorial(object):

    def __init__(self, dim):
        assert dim >= 3, "dim may be no less than 3"
        self.dim = dim
        self.npts = 2 ** dim

    def generate_points(self):
        return 0.5*(1 + pydoe.ff2n(self.dim))


class BoxBehnken(object):

    def __init__(self, dim):
        self.dim = dim
        self.npts = pydoe.bbdesign(self.dim, center=1).shape[0]

    def generate_points(self):
        return 0.5*(1 + pydoe.bbdesign(self.dim, center=1))


# ========================= For Test =======================


def _main():
    print("========================= LHD =======================")
    lhs = LatinHypercube(4, 10, criterion='c')
    print(lhs.generate_points())

    print("\n========================= SLHD =======================")
    slhd = SymmetricLatinHypercube(3, 10)
    print(slhd.generate_points())

    print("\n========================= 2-Factorial =======================")
    twofact = TwoFactorial(3)
    print(twofact.generate_points())
    print(twofact.npts)

    print("\n========================= Box-Behnken =======================")
    bb = BoxBehnken(3)
    print(bb.generate_points())
    print(bb.npts)

if __name__ == "__main__":
    _main()



"""
.. module:: test_multisampling
  :synopsis: Test multisampling strategy
.. moduleauthor:: David Eriksson <dme65@cornell.edu>
"""

import logging
from pySOT import *
from poap.controller import SerialController
import numpy as np
import os.path


def main():
    if not os.path.exists("./logfiles"):
        os.makedirs("logfiles")
    if os.path.exists("./logfiles/test_multisearch.log"):
        os.remove("./logfiles/test_multisearch.log")
    logging.basicConfig(filename="./logfiles/test_multisearch.log",
                        level=logging.INFO)

    print("\nNumber of threads: 1")
    print("Maximum number of evaluations: 500")
    print("Search strategy: CandidateDYCORS, Genetic Algorithm, Multi-Start Gradient")
    print("Experimental design: Latin Hypercube")
    print("Surrogate: Cubic RBF")

    nthreads = 1
    maxeval = 500
    nsamples = nthreads

    data = Ackley(dim=10)
    print(data.info)

    # Create a strategy and a controller
    sampling_method = [CandidateDYCORS(data=data, numcand=100*data.dim),
                       GeneticAlgorithm(data=data), MultiStartGradient(data=data)]
    controller = SerialController(data.objfunction)
    controller.strategy = \
        SyncStrategyNoConstraints(
            worker_id=0, data=data,
            maxeval=maxeval, nsamples=nsamples,
            response_surface=RBFInterpolant(surftype=CubicRBFSurface, maxp=maxeval),
            exp_design=SymmetricLatinHypercube(dim=data.dim, npts=2*(data.dim + 1)),
            sampling_method=MultiSampling(sampling_method, [0, 1, 2]))

    result = controller.run()
    best, xbest = result.value, result.params[0]

    print('Best value: {0}'.format(best))
    print('Best solution: {0}\n'.format(
        np.array_str(xbest, max_line_width=np.inf,
                     precision=5, suppress_small=True)))


if __name__ == '__main__':
    main()



import numpy as np


class Keane:
    def __init__(self, dim=10):
        self.xlow = np.zeros(dim)
        self.xup = 5 * np.ones(dim)
        self.dim = dim
        self.min = -0.835
        self.integer = []
        self.continuous = np.arange(0, dim)
        self.info = str(dim)+"-dimensional Keane bump function \n" +\
                             "Global optimum: -0.835 for large n"

    # Return a list with the constraint function at the given point
    def eval_ineq_constraints(self, x):
        vec = np.zeros((x.shape[0], 2))
        vec[:, 0] = 0.75 - np.prod(x)
        vec[:, 1] = np.sum(x) - 7.5 * self.dim
        return vec

    def deriv_ineq_constraints(self, x):
        vec = np.zeros((x.shape[0], 2, x.shape[1]))
        for i in range(x.shape[0]):
            xx = x[i, :]
            for j in range(x.shape[1]):
                vec[i, 0, j] = -np.prod(np.hstack((xx[:j], xx[j+1:])))
                vec[i, 1, j] = 1.0
        return vec

    # Evaluate the objective function for a single data point
    def objfunction(self, x):
        """Evaluate the Keane function  at x

        :param x: Data point
        :return: Value at x
        """
        n = len(x)
        return -abs((sum(np.cos(x)**4)-2 * np.prod(np.cos(x)**2)) /
                    max([1E-10, np.sqrt(np.dot(1+np.arange(n), x**2))]))



"""
.. module:: rbf_interpolant
   :synopsis: Basic radial basis function interpolation
.. moduleauthor:: David Bindel <bindel@cornell.edu>

:Module: rbf_interpolant
:Author: David Bindel <bindel@cornell.edu>
"""

import numpy as np
import numpy.linalg as la
import scipy.spatial as scp
from rbf_surfaces import CubicRBFSurface


class RBFInterpolant(object):
    """Compute and evaluate RBF interpolant.

    Manages an expansion of the form

    .. math::
        f(x) = \\sum_j c_j \\phi(\\|x-x_j\\|) + \\sum_j \\lambda_j p_j(x)

    where the functions :math:`p_j(x)` are low-degree polynomials.
    The fitting equations are

    .. math::

        \\begin{bmatrix} \\eta I & P^T \\\\ P & \\Phi+\\eta I \\end{bmatrix}
        \\begin{bmatrix} \\lambda \\\\ c \\end{bmatrix} =
        \\begin{bmatrix} 0 \\\\ f \\end{bmatrix}

    where :math:`P_{ij} = p_j(x_i)` and
    :math:`\\Phi_{ij}=\\phi(\\|x_i-x_j\\|)`.
    The regularization parameter :math:`\\eta` allows us to avoid problems
    with potential poor conditioning of the system.

    :ivar phi: Kernel function
    :ivar P: Tail functions
    :ivar dphi: Derivative of kernel function
    :ivar dP: Gradient of tail functions
    :ivar ntail: Number of tail functions
    :ivar nump: Current number of points
    :ivar maxp: Initial maximum number of points (can grow)
    :ivar A: Interpolation system matrix
    :ivar rhs: Right hand side for interpolation system
    :ivar x: Interpolation points
    :ivar fx: Values at interpolation points
    :ivar c: Expansion coefficients
    :ivar dim: Number of dimensions
    :ivar ntail: Number of tail functions
    """

    def __init__(self, surftype=CubicRBFSurface, maxp=100):
        self.maxp = maxp
        self.surface = None
        self.surftype = surftype

    @property
    def nump(self):
        return 0 if self.surface is None else self.surface.npt

    @property
    def x(self):
        return self.surface.x

    @property
    def fx(self):
        return self.surface.fx

    def reset(self):
        """Re-set the interpolation."""
        self.surface = None

    def get_x(self):
        """Get the list of data points

        :return: List of data points
        """
        return self.x

    def get_fx(self):
        """Get the list of function values for the data points.

        :return: List of function values
        """
        return np.expand_dims(self.fx, axis=1)

    def add_point(self, xx, fx):
        """Add a new function evaluation

        :param xx: Point to add
        :param fx: The function value of the point to add
        """
        if self.surface is None:
            self.surface = self.surftype(dim=len(xx), maxpts=self.maxp)
        else:
            assert xx.shape == (self.surface.dim,)
        self.surface.add_points(np.array([xx]), np.array([fx]))

    def transform_fx(self, fx):
        """Replace f with transformed function values for the fitting

        :param fx: Transformed function value
        """
        assert self.surface.npt == fx.shape[0]
        self.surface.set_fx(fx.ravel())

    def eval(self, xx, d=None):
        """Evaluate the rbf interpolant at the point xx

        :param xx: Point where to evaluate
        :return: Value of the rbf interpolant at x
        """
        assert xx.shape == (self.surface.dim,)
        return self.surface.eval(xx, d)

    def evals(self, xx, d=None):
        """Evaluate the rbf interpolant at the points xx

        :param xx: Points where to evaluate
        :return: Values of the rbf interpolant at x
        """
        assert xx.shape[1] == self.surface.dim
        return np.expand_dims(self.surface.eval(xx, d), axis=1)

    def deriv(self, x, d=None):
        """Evaluate the derivative of the rbf interpolant at x

        :param x: Data point
        :return: Derivative of the rbf interpolant at x
        """
        assert x.shape == (self.surface.dim,)
        return self.surface.deriv(x, d)


# ====================================================================


def _main():
    """Main test routine"""

    def test_f(x):
        """Test function"""
        fx = x[1]*np.sin(x[0]) + x[0]*np.cos(x[1])
        return fx

    def test_df(x):
        """Derivative of test function"""
        dfx = np.array([x[1]*np.cos(x[0])+np.cos(x[1]),
                        np.sin(x[0])-x[0]*np.sin(x[1])])
        return dfx

    fhat = RBFInterpolant(CubicRBFSurface, 1e-8, 20)
    xs = np.random.rand(120, 2)
    for i in range(100):
        xx = xs[i, :]
        fx = test_f(xx)
        fhat.add_point(xx, fx)
    fhx = fhat.evals(xs[:5, :])
    for i in range(5):
        fx = test_f(xs[i, :])
        print("Err: %e" % (abs(fx-fhx[i])/abs(fx)))
    for i in range(10):
        xx = xs[100+i, :]
        fx = test_f(xx)
        dfx = test_df(xx)
        fhx = fhat.eval(xx)
        dfhx = fhat.deriv(xx)
        print("Err (interp): %e : %e" % (abs(fx-fhx)/abs(fx),
                                         la.norm(dfx-dfhx)/la.norm(dfx)))

if __name__ == "__main__":
    _main()



from poap.controller import ProcessWorkerThread
import numpy as np
from subprocess import Popen, PIPE
import os

def array2str(x):
    return ",".join(np.char.mod('%f', x))

class SphereExt:
    def __init__(self, dim=10):
        self.xlow = -15 * np.ones(dim)
        self.xup = 20 * np.ones(dim)
        self.dim = dim
        self.info = str(dim)+"-dimensional Sphere function \n" +\
                             "Global optimum: f(0,0,...,0) = 0"
        self.min = 0
        self.integer = []
        self.continuous = np.arange(0, dim)
        assert os.path.isfile("./test/sphere_ext"), "You need to build sphere_ext" \
            " or specify another path and/or filename"


class objfunction(ProcessWorkerThread):

    def handle_eval(self, record):
        self.process = Popen(['./test/sphere_ext', array2str(record.params[0])],
                             stdout=PIPE)
        out = self.process.communicate()[0]
        try:
            val = float(out)  # This raises ValueError if out is not a float
            self.finish_success(record, val)
        except ValueError:
            self.finish_failure(record)


"""
.. module:: test_simple
  :synopsis: Test Simple
.. moduleauthor:: David Eriksson <dme65@cornell.edu>
"""

from pySOT import *
from poap.controller import ThreadController, BasicWorkerThread
import numpy as np
import os.path

class AckleyUnit:
    def __init__(self, dim=10):
        self.xlow = -1 * np.ones(dim)
        self.xup = 1 * np.ones(dim)
        self.dim = dim
        self.info = str(dim)+"-dimensional Ackley function on the unit sphere \n" +\
                             "Global optimum: f(1,0,...,0) = ... = f(0,0,...,1) = " +\
                             str(np.round(20*(1-np.exp(-0.2/np.sqrt(dim))), 3))
        self.min = 20*(1 - np.exp(-0.2/np.sqrt(dim)))
        self.integer = []
        self.continuous = np.arange(0, dim)
        check_opt_prob(self)

    def objfunction(self, x):
        n = float(len(x))
        return -20.0 * np.exp(-0.2*np.sqrt(np.sum(x**2)/n)) - \
            np.exp(np.sum(np.cos(2.0*np.pi*x))/n) + 20 + np.exp(1)

    def eval_eq_constraints(self, x):
        return np.linalg.norm(x) - 1


def main():
    if not os.path.exists("./logfiles"):
        os.makedirs("logfiles")
    if os.path.exists("./logfiles/test_projection.log"):
        os.remove("./logfiles/test_projection.log")
    logging.basicConfig(filename="./logfiles/test_projection.log",
                        level=logging.INFO)

    print("\nNumber of threads: 4")
    print("Maximum number of evaluations: 1000")
    print("Search strategy: CandidateDYCORS")
    print("Experimental design: Latin Hypercube")
    print("Surrogateurrogates: Cubic RBF")

    nthreads = 4
    maxeval = 1000
    nsamples = nthreads

    data = AckleyUnit(dim=10)
    print(data.info)

    def projection(x):
        return x / np.linalg.norm(x)

    # Create a strategy and a controller
    controller = ThreadController()
    controller.strategy = \
        SyncStrategyProjection(
            worker_id=0, data=data,
            maxeval=maxeval, nsamples=nsamples,
            exp_design=LatinHypercube(dim=data.dim, npts=2*(data.dim+1)),
            response_surface=RBFInterpolant(surftype=CubicRBFSurface, maxp=maxeval),
            sampling_method=CandidateDYCORS(data=data, numcand=100*data.dim),
            proj_fun=projection
        )

    # Launch the threads and give them access to the objective function
    for _ in range(nthreads):
        worker = BasicWorkerThread(controller, data.objfunction)
        controller.launch_worker(worker)

    # Run the optimization strategy
    result = controller.run()

    print('Best value found: {0}'.format(result.value))
    print('Best solution found: {0}'.format(
        np.array_str(result.params[0], max_line_width=np.inf,
                     precision=5, suppress_small=True)))
    print('||x||_2 = {0}\n'.format(np.linalg.norm(result.params[0])))


if __name__ == '__main__':
    main()



"""
.. module:: test_ensemble
  :synopsis: Test Ensemble surrogates
.. moduleauthor:: David Eriksson <dme65@cornell.edu>
"""

import logging
from pySOT import *
from poap.controller import ThreadController, BasicWorkerThread
import numpy as np
import os.path


def main():
    if not os.path.exists("./logfiles"):
        os.makedirs("logfiles")
    if os.path.exists("./logfiles/test_ensemble.log"):
        os.remove("./logfiles/test_ensemble.log")
    logging.basicConfig(filename="./logfiles/test_ensemble.log",
                        level=logging.INFO)

    print("\nNumber of threads: 5")
    print("Maximum number of evaluations: 250")
    print("Search strategy: CandidateSRBF")
    print("Experimental design: Latin Hypercube + point [1,1,...,1]")
    print("Ensemble Surrogate: Cubic RBF, PolyReg")

    nthreads = 5
    maxeval = 250
    nsamples = nthreads

    data = Ackley(dim=5)
    print(data.info)

    # Use RBF + PolyReg
    bounds = np.vstack((data.xlow, data.xup)).T
    basisp = basis_TD(data.dim, 2)  # use order 2 and no cross-terms

    models = [
        RBFInterpolant(surftype=CubicRBFSurface, maxp=maxeval),
        PolyRegression(bounds, basisp)
    ]
    response_surface = EnsembleSurrogate(models, maxeval)

    # Add an additional point to the experimental design. If a good
    # solution is already known you can add this point to the
    # experimental design
    extra = np.ones((1, data.dim))

    # Create a strategy and a controller
    controller = ThreadController()
    controller.strategy = \
        SyncStrategyNoConstraints(
            worker_id=0, data=data,
            response_surface=response_surface,
            maxeval=maxeval, nsamples=nsamples,
            exp_design=SymmetricLatinHypercube(dim=data.dim, npts=2*(data.dim+1)),
            sampling_method=CandidateSRBF(data=data, numcand=100*data.dim),
            extra=extra)

    # Launch the threads and give them access to the objective function
    for _ in range(nthreads):
        worker = BasicWorkerThread(controller, data.objfunction)
        controller.launch_worker(worker)

    # Run the optimization strategy
    result = controller.run()

    response_surface.compute_weights()
    print('Final weights: {0}'.format(
        np.array_str(response_surface.weights, max_line_width=np.inf,
                     precision=5, suppress_small=True)))

    print('Best value found: {0}'.format(result.value))
    print('Best solution found: {0}\n'.format(
        np.array_str(result.params[0], max_line_width=np.inf,
                     precision=5, suppress_small=True)))


if __name__ == '__main__':
    main()



import numpy as np
import matplotlib.pyplot as plt


def to_unit_box(x, data):
    return (np.copy(x) - data.xlow) / (data.xup - data.xlow)


def from_unit_box(x, data):
        return data.xlow + (data.xup - data.xlow) * np.copy(x)


def unit_rescale(xx):
    """Shift and rescale elements of a vector to the unit interval
    """
    xmax = np.amax(xx)
    xmin = np.amin(xx)
    if xmax == xmin:
        return np.ones(xx.shape)
    else:
        return (xx-xmin)/(xmax-xmin)


def round_vars(data, x):
    """Round integer variables to closest integer
    """
    if len(data.integer) > 0:
        # Round the original ranged integer variables
        x[:, data.integer] = np.round(x[:, data.integer])
        # Make sure we don't violate the bound constraints
        for i in data.integer:
            ind = np.where(x[:, i] < data.xlow[i])
            x[ind, i] += 1
            ind = np.where(x[:, i] > data.xup[i])
            x[ind, i] -= 1
    return x


def check_opt_prob(obj):
    """Routine for checking that an implementation of the optimization problem
    follows the standard. This method checks everything, but can't make
    sure that the objective function and constraint methods return values
    of the correct type since this would involve actually evaluating the
    objective function which isn't feasible when the evaluations are
    expensive. If some test fails, an exception is raised through assert.

    :param obj: Objective function
    """
    assert hasattr(obj, "dim"), \
        "Problem dimension required"
    assert hasattr(obj, "xlow"), \
        "Numpy array of lower bounds required"
    assert isinstance(obj.xlow, np.ndarray), \
        "Numpy array of lower bounds required"
    assert hasattr(obj, "xup"), \
        "Numpy array of upper bounds required"
    assert isinstance(obj.xup, np.ndarray), \
        "Numpy array of upper bounds required"
    assert hasattr(obj, "integer"), \
        "Integer variables must be specified"
    if len(obj.integer) > 0:
        assert isinstance(obj.integer, np.ndarray), \
            "Integer variables must be specified"
    else:
        assert isinstance(obj.integer, np.ndarray) or \
            isinstance(obj.integer, list), \
            "Integer variables must be specified"
    assert hasattr(obj, "continuous"), \
        "Continuous variables must be specified"
    if len(obj.continuous) > 0:
        assert isinstance(obj.continuous, np.ndarray), \
            "Continuous variables must be specified"
    else:
        assert isinstance(obj.continuous, np.ndarray) or \
            isinstance(obj.continuous, list), \
            "Continuous variables must be specified"

    # Check for logical errors
    assert isinstance(obj.dim, int) and obj.dim > 0, \
        "Problem dimension must be a positive integer."
    assert (len(obj.xlow) == obj.dim and
            len(obj.xup) == obj.dim), \
        "Incorrect size for xlow and xup"
    assert all(obj.xlow[i] < obj.xup[i] for i in range(obj.dim)), \
        "Lower bounds must be below upper bounds."
    if len(obj.integer) > 0:
        assert np.amax(obj.integer) < obj.dim and np.amin(obj.integer) >= 0, \
            "Integer variable index can't exceed " \
            "number of dimensions or be negative"
    if len(obj.continuous) > 0:
        assert np.amax(obj.continuous) < obj.dim and \
               np.amin(obj.continuous) >= 0, \
               "Continuous variable index can't exceed " \
               "number of dimensions or be negative"
    assert len(np.intersect1d(obj.continuous, obj.integer)) == 0, \
        "A variable can't be both an integer and continuous"
    assert len(obj.continuous)+len(obj.integer) == obj.dim, \
        "All variables must be either integer or continuous"


def progress_plot(controller, title=''):

    # Extract function values from the controller, ignoring crashed evaluations
    fvals = np.array([o.value for o in controller.fevals if o.value is not None])

    plt.ion()
    plt.figure()
    plt.plot(np.arange(0, fvals.shape[0]), fvals, 'bo')  # Points
    plt.plot(np.arange(0, fvals.shape[0]), np.minimum.accumulate(fvals),
             'r-', linewidth=4.0)  # Best value found

    # Set limits
    ymin = np.min(fvals)
    ymax = np.max(fvals)
    plt.ylim(ymin - 0.1 * (ymax - ymin), ymax + 0.1 * (ymax - ymin))

    plt.xlabel('Evaluations')
    plt.ylabel('Function Value')
    plt.title(title)
    plt.show()
